"""
Build friction or barrier layers from raster and vector data.
"""
import logging
from pathlib import Path
from typing import Dict
from warnings import warn

import numpy as np
import numpy.typing as npt

from reVX.handlers.layered_h5 import LayeredTransmissionH5
from reVX.config.transmission_layer_creation import (Extents,
                                                     LayerBuildConfig,
                                                     LayerBuildComponents)
from reVX.least_cost_xmission.layers.base import BaseLayerCreator
from reVX.least_cost_xmission.layers.utils import rasterize_shape_file
from reVX.least_cost_xmission.layers.masks import MaskArr, Masks
from reVX.least_cost_xmission.config.constants import (DEFAULT_DTYPE, ALL,
                                                       METERS_IN_MILE,
                                                       CELL_SIZE)

logger = logging.getLogger(__name__)
TIFF_EXTENSIONS = {'.tif', '.tiff'}
SHP_EXTENSIONS = {'.shp', '.gpkg'}


class LayerCreator(BaseLayerCreator):
    """Build layer based on tiff and user config."""

    def __init__(self, io_handler: LayeredTransmissionH5,
                 masks: Masks, output_tiff_dir=".",
                 dtype: npt.DTypeLike=DEFAULT_DTYPE,
                 cell_size=CELL_SIZE):
        """
        Parameters
        ----------
        io_handler : :class:`LayeredTransmissionH5`
            Transmission layer IO handler
        masks : Masks
            Masks instance that can be used to retrieve multiple types
            of masks.
        output_tiff_dir : path-like, optional
            Directory where cost layers should be saved as GeoTIFF.
            By default, ``"."``.
        dtype : np.dtype, optional
            Data type for final dataset. By default, ``float32``.
        cell_size : int, optional
            Side length of each cell, in meters. Cells are assumed to be
            square. By default, :obj:`CELL_SIZE`.
        """
        self._masks = masks
        super().__init__(io_handler=io_handler, mask=None,
                         output_tiff_dir=output_tiff_dir, dtype=dtype,
                         cell_size=cell_size)

    def build(self, layer_name, build_config: LayerBuildComponents,
              values_are_costs_per_mile=False, write_to_h5=True,
              description=None):
        """
        Combine multiple GeoTIFFs and vectors to a raster layer and save
        to GeoTIFF/HDF5 file.

        Parameters
        ----------
        layer_name : str
            Name of layer to use in H5 and for output tiff.
        build_config : LayerBuildComponents
            Dict of LayerBuildConfig keyed by GeoTIFF/vector filenames.
        values_are_costs_per_mile : bool, default=False
            Option to convert values into costs per cell under the
            assumption that the resulting values arte costs in $/mile.
            By default, ``False``, which writes raw values to TIFF/H5.
        write_to_h5 : bool, default=True
            Option to write the layer to H5 file after creation.
            By default, ``True``.
        description : str, optional
            Optional description to store with this layer in the H5
            file. By default, ``None``.
        """
        layer_name = layer_name.replace(".tif", "").replace(".tiff", "")
        logger.debug('Combining %s layers', layer_name)
        result = np.zeros(self._io_handler.shape, dtype=self._dtype)
        fi_layers: LayerBuildComponents = {}

        for fname, config in build_config.items():
            if config.forced_inclusion:
                fi_layers[fname] = config
                continue

            logger.debug('Processing %s with config %s', fname, config)
            if Path(fname).suffix.lower() in TIFF_EXTENSIONS:
                data = self._io_handler.load_data_using_h5_profile(
                    fname, reproject=True)
                temp = self._process_raster_layer(data, config)
                result += temp
            elif Path(fname).suffix.lower() in SHP_EXTENSIONS:
                temp = self._process_vector_layer(fname, config)
                result += temp
            else:
                raise ValueError(f'Unsupported file extension on {fname}')

        result = self._process_forced_inclusions(result, fi_layers)
        if values_are_costs_per_mile:
            result = result / METERS_IN_MILE * self._cell_size

        out_filename = self.output_tiff_dir / f"{layer_name}.tif"
        logger.debug('Writing combined %s layers to %s', layer_name,
                     out_filename)
        self._io_handler.save_data_using_h5_profile(result, out_filename)
        if write_to_h5:
            out = self._io_handler.load_data_using_h5_profile(
                out_filename, reproject=True)
            logger.debug('Writing %s to H5', layer_name)
            self._io_handler.write_layer_to_h5(out, layer_name,
                                               description=description)

    def _process_raster_layer(self, data: npt.NDArray,  # type: ignore[return]
                              config: LayerBuildConfig) -> npt.NDArray:
        """Create the desired layer from the array using LayerBuildConfig.

        Desired "range" or "map" operation is only applied to the area
        indicated by the "extent".

        Parameters
        ----------
        data : array-like
            Array of data to process.
        config : LayerBuildConfig
            Definition of layer processing.

        Returns
        -------
        array-like | None
            Transformed data.
        """
        self._check_tiff_layer_config(config)

        if config.global_value is not None:
            temp = np.full(self._io_handler.shape,
                           fill_value=config.global_value,
                           dtype=self._dtype)

            if config.extent == ALL:
                return temp

            mask = self._get_mask(config.extent)
            processed = np.zeros(self._io_handler.shape, dtype=self._dtype)
            processed[mask] = temp[mask]
            return processed

        # Assign all cells one or more ranges to a value
        if config.bins is not None:
            _validate_bin_range(config.bins)
            _validate_bin_continuity(config.bins)

            processed = np.zeros(self._io_handler.shape, dtype=self._dtype)
            if config.extent != ALL:
                mask = self._get_mask(config.extent)

            for i, interval in enumerate(config.bins):
                logger.debug('Calculating layer values for bin '
                             f'{i+1}/{len(config.bins)}: {interval}')
                min, max = interval.min, interval.max
                temp = np.where(np.logical_and(data >= min, data < max),
                                interval.value, 0)

                if config.extent == ALL:
                    processed += temp
                    continue

                processed[mask] += temp[mask]

            return processed

        if config.pass_through:
            if config.extent == ALL:
                return data

            mask = self._get_mask(config.extent)
            processed = np.zeros(self._io_handler.shape, dtype=self._dtype)
            processed[mask] = data[mask]
            return processed

        # No bins specified, has to be map. Assign cells values based on map.
        temp = np.zeros(self._io_handler.shape, dtype=self._dtype)
        for key, val in config.map.items():  # type: ignore[union-attr]
            temp[data == key] = val

        if config.extent == ALL:
            return temp

        mask = self._get_mask(config.extent)
        processed = np.zeros(self._io_handler.shape, dtype=self._dtype)
        processed[mask] = temp[mask]
        return processed

    def _process_vector_layer(self, fname: str, config: LayerBuildConfig
                              ) -> npt.NDArray:
        """
        Rasterize a vector layer

        Parameters
        ----------
        fname : str
            Name of vector layer to rasterize
        config : LayerBuildConfig
            Config for layer

        Returns
        -------
        array-like
            Rasterized vector
        """
        if config.rasterize is None:
            raise ValueError(f'{fname} is a vector but the config is missing '
                             f'key "rasterize": {config}')

        r_config = config.rasterize

        temp = rasterize_shape_file(fname, self._io_handler.profile,
                                    buffer_dist=r_config.buffer,
                                    burn_value=r_config.value,
                                    dtype=self._dtype,
                                    reproject_vector=r_config.reproject)

        if config.extent == ALL:
            return temp

        mask = self._get_mask(config.extent)
        processed = np.zeros(self._io_handler.shape, dtype=self._dtype)
        processed[mask] = temp[mask]
        return processed

    def _process_forced_inclusions(self, data: npt.NDArray,
                                   fi_layers: LayerBuildComponents
                                   ) -> npt.NDArray:
        """
        Use forced inclusion (FI) layers to remove barriers/friction. Any
        value > 0 in the FI layers will result in a 0 in the corresponding
        cell in the returned raster.

        Parameters
        ----------
        data : array-like
            Composite friction or barrier layer
        fi_layers : LayerBuildComponents
            Dict of forced inclusions layers keyed by GeoTIFF filename.

        Returns
        -------
        array-like
            Composite layer with forced inclusions added
        """
        fi = np.zeros(self._io_handler.shape)

        for fname, config in fi_layers.items():
            if Path(fname).suffix.lower() not in ['.tif', '.tiff']:
                msg = (f'Forced inclusion file {fname} does not end with .tif.'
                       ' GeoTIFFs are the only format allowed for forced '
                       'inclusions.')
                logger.error(msg)
                raise ValueError(msg)

            global_value_given = config.global_value is not None
            map_given = config.map is not None
            range_given = config.bins is not None
            rasterize_given = config.rasterize is not None
            bad_input_given = (global_value_given
                               or map_given
                               or range_given
                               or rasterize_given)
            if bad_input_given:
                msg = ('`global_value`, `map`, `bins`, and `rasterize` are '
                       'not allowed if `forced_inclusion` is True, but one '
                       f'was found in config: {fname}: {config}')
                logger.error(msg)
                raise ValueError(msg)

            # Past guard clauses, process FI
            if config.extent != ALL:
                mask = self._get_mask(config.extent)

            temp = self._io_handler.load_data_using_h5_profile(fname,
                                                               reproject=True)

            if config.extent == ALL:
                fi += temp
            else:
                fi[mask] += temp[mask]

        data[fi > 0] = 0
        return data

    def _get_mask(self, extent: Extents) -> MaskArr:
        """
        Get mask by requested extent

        Parameters
        ----------
        extent : Extents
            Extent of desired mask, 'all' is not allowed.

        Returns
        -------
        MaskArr
            Mask array
        """
        if extent == ALL:
            raise AttributeError(f'Mask for extent of {extent} is unnecessary')

        if extent == 'wet':
            mask = self._masks.wet_mask
        elif extent == 'wet+':
            mask = self._masks.wet_plus_mask
        elif extent == 'dry':
            mask = self._masks.dry_mask
        elif extent == 'dry+':
            mask = self._masks.dry_plus_mask
        elif extent == 'landfall':
            mask = self._masks.landfall_mask
        else:
            raise AttributeError(f'Unknown mask type: {extent}')

        return mask

    @staticmethod
    def _check_tiff_layer_config(config: LayerBuildConfig):
        """ Check if a LayerBuildConfig is valid for a GeoTIFF.

        Parameters
        ----------
        config : LayerBuildConfig
            The config model to check

        Raises
        ------
        ValueError
            If an issue is detected
        """
        if config.rasterize is not None:
            raise ValueError('"rasterize" is only for vectors. Found in '
                             f'config {config}')

        mutex_entries = [config.map, config.bins, config.global_value]
        num_entries = sum(entry is not None for entry in mutex_entries)
        num_entries += int(config.pass_through)
        if num_entries > 1:
            raise ValueError('Keys "global_value", "map", "bins", and '
                             '"pass_through" are mutually exclusive but '
                             'more than one was found in raster config '
                             f'{config}')

        if num_entries < 1:
            raise ValueError('Either "global_value", "map", "bins", and '
                             '"pass_through" must be specified fora raster, '
                             'but none were found in config '
                             f'{config}')


def _validate_bin_range(bins):
    """Check for correctness in bin range. """
    for bin in bins:
        if bin.min > bin.max:
            raise AttributeError('Min is greater than max for bin config '
                                 f'{bin}.')
        if bin.min == float('-inf') and bin.max == float('inf'):
            msg = ('Bin covers all possible values, did you forget to set '
                   f'min or max? {bin}')
            logger.warning(msg)
            warn(msg)


def _validate_bin_continuity(bins):
    """Warn user of potential gaps in bin range continuity."""
    sorted_bins = sorted(bins, key=lambda x: x.min)
    last_max = float('-inf')
    for i, bin in enumerate(sorted_bins):
        if bin.min < last_max:
            last_bin = sorted_bins[i - 1] if i > 0 else '-infinity'
            msg = (f'Overlapping bins detected between bin {last_bin} '
                   f'and {bin}')
            logger.warning(msg)
            warn(msg)
        if bin.min > last_max:
            last_bin = sorted_bins[i - 1] if i > 0 else '-infinity'
            msg = f'Gap detected between bin {last_bin} and {bin}'
            logger.warning(msg)
            warn(msg)
        if i + 1 == len(sorted_bins):
            if bin.max < float('inf'):
                msg = f'Gap detected between bin {bin} and infinity'
                logger.warning(msg)
                warn(msg)

        last_max = bin.max



"""
Standard filenames, H5 layer names, etc.
"""

ALL = 'all'

ISO_H5_LAYER_NAME = 'ISO_regions'
BARRIER_H5_LAYER_NAME = 'transmission_barrier'  # Combined friction and barrier
DRY_MULTIPLIER_TIFF = 'dry_multipliers.tif'

DEFAULT_DTYPE = 'float32'


METERS_IN_MILE = 1609.344
"""1609.344m : Meters in a mile - useful for conversions. """
SHORT_MULT = 1.5
"""1.5: Short-length spur line multiplier."""
MEDIUM_MULT = 1.2
"""1.2: Medium-length spur line multiplier."""
SHORT_CUTOFF = 3 * METERS_IN_MILE / 1000
"""4.8 km (3 miles): Tie line length below which ``SHORT_MULT`` is applied."""
MEDIUM_CUTOFF = 10 * METERS_IN_MILE / 1000
"""16 km (10 miles): Tie line length below which ``MEDIUM_MULT`` is applied"""

CELL_SIZE = 90
"""90 m: Standard size of reV exclusions cell. Both dims must be equal"""

RESOLUTION = 128
"""128: Standard reV supply curve cell resolution. """

CLIP_RASTER_BUFFER = 1.05
"""1.05: Decimal % distance to buffer clipped cost raster by.

This helps to find the cheapest path. Larger values will run slower.
Should be greater than 1 or clip will be too small.
"""

NUM_NN_SINKS = 2
"""2: Default number of nearest neighbor sinks to use for clipping radius. """

BARRIERS_MULT = 100
"""100: Default barrier multipier value.

Costs multiplier for cells affected by transmission barriers This should
be greater than the largest possible cost multiplier * number of cells
to discourage from crossing"""

MINIMUM_SPUR_DIST_KM = 0
"""0 km: Minimum spur line path length.

Anything less will be set to this value. """

SINK_CONNECTION_COST = 1e11
"""$1e11: Cost to connect to PCA load center.

Completely synthetic and should only be used if no other connection is
available. """

# Default slope multipliers and cutoffs
HILL_MULT = 1
"""1: Default hill slope multiplier value. """
MTN_MULT = 1
"""1: Default mountain slope multiplier value. """
HILL_SLOPE = 2
"""2: Default hill slope cutoff value.

Slope values above this (inclusive) are considered hills. """
MTN_SLOPE = 8
"""8: Default mountain slope cutoff value.

Slope values above this (inclusive) are considered mountains. """

WATER_NLCD_CODE = 11
"""11: NLCD catergory value for warter. """
WATER_MULT = 10.0
"""10: Multiplier value for water cells based on NLCD. """

TEST_DEFAULT_MULTS = {
    "iso": "default",
    "land_use": {
        "cropland": 1,
        "forest": 4,
        "suburban": 5,
        "urban": 6,
        "wetland": 7
    },
    "slope": {
        "hill_mult": 2,
        "hill_slope": 2,
        "mtn_mult": 5,
        "mtn_slope": 8
    }
}
"""dict: Default test multipliers for RI"""

# Category strings for transmission features
TRANS_LINE_CAT = 'TransLine'
"""'TransLine': Transmission line feature category label. """
LOAD_CENTER_CAT = 'LoadCen'
"""'LoadCen': load center feature category label. """
SINK_CAT = 'PCALoadCen'
"""'PCALoadCen': Sink feature category label. """
SUBSTATION_CAT = 'Substation'
"""'Substation': Substation feature category label. """




# -*- coding: utf-8 -*-
"""
reVX Utilities
"""
import addfips
import geopandas as gpd
import pandas as pd
import numpy as np
import pyproj
import rasterio
import shapely
import skimage
import sklearn
from sklearn.metrics.pairwise import haversine_distances

from reV.utilities import log_versions as reV_log_versions
from reVX.version import __version__


def coordinate_distance(coords1, coords2):
    """
    Compute the haversine distance between the two sets of coordinates.
    Results are in km

    Parameters
    ----------
    coords1 : ndarray
        First set of (lat, lon) coordinates
    coords2 : ndarray
        Second set of (lat, lon) coordinates

    Returns
    -------
    dist : ndarray
        Vector of haversine distances between coordinate set 1 and set 2 in km
    """
    dist = haversine_distances(np.radians(coords1), np.radians(coords2))
    if len(coords1) == 1:
        dist = dist.ravel()
    else:
        dist = np.diag(dist)

    # radius of the earth in kilometers # radius of the earth in km
    R = 6371.0

    return dist * R


def log_versions(logger):
    """Log package versions:

        - rex, reV, and reVX to info
        - h5py, numpy, pandas, scipy, PySAM, geopandas, pyproj,
          rasterio, shapely, sklearn, and skimage to debug

    Parameters
    ----------
    logger : logging.Logger
        Logger object to log memory message to.
    """
    logger.info('Running with reVX version {}'.format(__version__))
    reV_log_versions(logger)
    logger.debug('- geopandas version {}'.format(gpd.__version__))
    logger.debug('- pyproj version {}'.format(pyproj.__version__))
    logger.debug('- rasterio version {}'.format(rasterio.__version__))
    logger.debug('- shapely version {}'.format(shapely.__version__))
    logger.debug('- scikit-image version {}'.format(skimage.__version__))
    logger.debug('- scikit-learn version {}'.format(sklearn.__version__))


def to_geo(data_frame, lat_col="latitude", lon_col="longitude",
           crs="epsg:4326"):
    """Convert a Pandas DataFrame to a GeoPandas GeoDataFrame.

    The input DataFrame must have latitude and longitude columns, which
    get converted to a point geometry in the outputs GeoDataFrame.

    Parameters
    ----------
    data_frame : pandas.DataFrame
        A pandas data frame with latitude and longitude coordinates.
    lat_col : str, optional
        The name of the latitude column. By default, ``"latitude"``.
    lon_col : str, optional
        The name of the longitude column. By default, ``"longitude"``.
    crs : str, optional
        The Coordinate Reference System of the output DataFrame
        represented as a string. By default, ``"epsg:4326"``.

    Returns
    -------
    geopandas.GeoDataFrame
        A GeoPandas GeoDataFrame object with points representing the
        lat/lon positions as the geometry.
    """
    missing = {col for col in [lat_col, lon_col] if col not in data_frame}
    if any(missing):
        raise KeyError("Input DataFrame missing the following required keys: "
                       "{}".format(missing))

    # pylint: disable=unnecessary-lambda-assignment
    to_point = lambda x: shapely.geometry.Point((x[lon_col], x[lat_col]))
    data_frame["geometry"] = data_frame.apply(to_point, axis=1)
    return gpd.GeoDataFrame(data_frame, geometry="geometry", crs=crs)


def load_fips_to_state_map():
    """Generate a FIPS to state name mapping.

    The keys of the returned dictionary are two-digit FIPS codes (as
    strings) and the values are the state names.

    Returns
    -------
    dict
        Dictionary mapping two-digitFIPS codes (as strings) to state
        names.
    """
    cdf = pd.read_csv(addfips.AddFIPS.data / "data" / "states.csv")
    cdf["fips"] = cdf["fips"].apply(lambda x: f"{x:02d}")
    return dict(zip(cdf["fips"], cdf["name"]))



# -*- coding: utf-8 -*-
"""
Base classes for setback exclusion computation
"""
import os
import logging
from copy import deepcopy
from warnings import warn
from math import floor, ceil
from itertools import product
from abc import abstractmethod
from concurrent.futures import as_completed

import numpy as np
import geopandas as gpd
from shapely.ops import unary_union
from shapely.validation import make_valid
from rasterio import (windows, transform, Affine, coords,
                      features as rio_features)

from rex.utilities import log_mem
from rex.utilities.execution import SpawnProcessPool
from reV.handlers.exclusions import ExclusionLayers
from reVX.handlers.geopackage import GPKGMeta
from reVX.utilities.exclusions import AbstractBaseExclusionsMerger
from reVX.setbacks.functions import (parcel_buffer, positive_buffer,
                                     features_clipped_to_county,
                                     features_with_centroid_in_county)

logger = logging.getLogger(__name__)


BUFFERS = {
    "default": positive_buffer,
    "parcel": parcel_buffer,
}
"""Types of buffers available for setback calculations. """


FEATURE_FILTERS = {
    "centroid": features_with_centroid_in_county,
    "clip": features_clipped_to_county,
}
"""Types of feature filters available for setback calculations. """


class Rasterizer:
    """Helper class to rasterize shapes."""

    def __init__(self, excl_fpath, weights_calculation_upscale_factor,
                 hsds=False):
        """
        Parameters
        ----------
        excl_fpath : str
            Path to .h5 file containing template layers. The raster will
            match the shape and profile of these layers.
        weights_calculation_upscale_factor : int
            If this value is an int > 1, the output will be a layer with
            **inclusion** weight values (floats ranging from 0 to 1).
            Note that this is backwards w.r.t the typical output of
            exclusion integer values (1 for excluded, 0 otherwise).
            Values <= 1 will still return a standard exclusion mask.
            For example, a cell that was previously excluded with a
            a boolean mask (value of 1) may instead be converted to an
            inclusion weight value of 0.75, meaning that 75% of the area
            corresponding to that point should be included (i.e. the
            exclusion feature only intersected a small portion - 25% -
            of the cell). This percentage inclusion value is calculated
            by upscaling the output array using this input value,
            rasterizing the exclusion features onto it, and counting the
            number of resulting sub-cells excluded by the feature. For
            example, setting the value to 3 would split each output
            cell into nine sub-cells - 3 divisions in each dimension.
            After the feature is rasterized on this high-resolution
            sub-grid, the area of the non-excluded sub-cells is totaled
            and divided by the area of the original cell to obtain the
            final inclusion percentage. Therefore, a larger upscale
            factor results in more accurate percentage values. If
            ``None`` (or a value <= 1), this process is skipped and the
            output is a boolean exclusion mask. By default ``None``.
        """
        props = _parse_excl_properties(excl_fpath, hsds=hsds)
        self._shape, self._profile = props
        self._scale_factor = weights_calculation_upscale_factor or 1
        self._scale_factor = int(self._scale_factor // 1)

    @property
    def profile(self):
        """Geotiff profile.

        Returns
        -------
        dict
        """
        return self._profile

    @property
    def transform(self):
        """rasterio.Affine: Affine transform for exclusion layer. """
        return Affine(*self.profile["transform"])

    @property
    def arr_shape(self):
        """Rasterize array shape.

        Returns
        -------
        tuple
        """
        return self._shape

    @property
    def inclusions(self):
        """Flag indicating wether or not the output raster represents
        inclusion values.

        Returns
        -------
        bool
        """
        return self._scale_factor > 1

    def _no_exclusions_array(self, multiplier=1, window=None):
        """Get an array of the correct shape representing no exclusions.

        The array contains all zeros, and a new one is created
        for every function call.

        Parameters
        ----------
        multiplier : int, optional
            Integer multiplier value used to scale up the dimensions of
            the array exclusions array (e.g. multiplier of 3 turns an
            array of shape (10, 20) into an array of shape (30, 60)).
        window : :cls:`rasterio.windows.Window`
            A ``rasterio`` window defining the area of the raster. Can
            be used to speed up computation and decrease memory
            requirements if features are localized to a small portion of
            the raster array.

        Returns
        -------
        np.array
            Array of zeros representing no exclusions.
        """
        if window is None:
            shape = tuple(x * multiplier for x in self.arr_shape[1:])
        else:
            shape = (window.height * multiplier, window.width * multiplier)
        return np.zeros(shape, dtype='uint8')

    def rasterize(self, shapes, window=None):
        """Convert geometries into exclusions array.

        Parameters
        ----------
        shapes : list, optional
            List of geometries to rasterize (i.e. list(gdf["geometry"])).
            If `None` or empty list, returns array of zeros.
        window : :obj:`rasterio.windows.Window`
            A ``rasterio`` window defining the area of the raster. Can
            be used to speed up computation and decrease memory
            requirements if features are localized to a small portion of
            the raster array.

        Returns
        -------
        arr : ndarray
            Rasterized array of shapes.
        """

        shapes = shapes or []
        shapes = [(geom, 1) for geom in shapes if geom is not None]

        if self.inclusions:
            return self._rasterize_to_weights(shapes, window)

        return self._rasterize_to_mask(shapes, window)

    def _rasterize_to_weights(self, shapes, window):
        """Rasterize features to weights using a high-resolution array."""

        if not shapes:
            return ((1 - self._no_exclusions_array(window=window))
                    .astype(np.float32))

        hr_arr = self._no_exclusions_array(multiplier=self._scale_factor,
                                           window=window)
        transform = self._window_transform(window)
        transform *= transform.scale(1 / self._scale_factor)

        rio_features.rasterize(shapes=shapes, out=hr_arr, fill=0,
                               transform=transform)

        arr = self._aggregate_high_res(hr_arr, window)
        return 1 - (arr / self._scale_factor ** 2)

    def _rasterize_to_mask(self, shapes, window):
        """Rasterize features with to an exclusion mask."""

        arr = self._no_exclusions_array(window=window)
        if shapes:
            transform = self._window_transform(window)
            rio_features.rasterize(shapes=shapes, out=arr, fill=0,
                                   transform=transform)

        return arr

    def _aggregate_high_res(self, hr_arr, window):
        """Aggregate the high resolution exclusions array to output shape. """

        arr = self._no_exclusions_array(window=window).astype(np.float32)
        for i, j in product(range(self._scale_factor),
                            range(self._scale_factor)):
            arr += hr_arr[i::self._scale_factor, j::self._scale_factor]
        return arr

    def _window_transform(self, window):
        """Calculate the transform for a given window, if any. """
        if window is None:
            return deepcopy(self.transform)
        return windows.transform(window, self.transform)


class AbstractBaseSetbacks(AbstractBaseExclusionsMerger):
    """Base class for Setbacks Calculators"""

    def __init__(self, excl_fpath, regulations, features, hsds=False,
                 weights_calculation_upscale_factor=None):
        """
        Parameters
        ----------
        excl_fpath : str
            Path to .h5 file containing exclusion layers, will also be
            the location of any new setback layers
        regulations : `~reVX.setbacks.regulations.SetbackRegulations`
            A `SetbackRegulations` object used to extract setback
            distances.
        features : str
            Path to file containing features to compute exclusions from.
        hsds : bool, optional
            Boolean flag to use h5pyd to handle .h5 'files' hosted on
            AWS behind HSDS. By default `False`.
        weights_calculation_upscale_factor : int, optional
            If this value is an int > 1, the output will be a layer with
            **inclusion** weight values instead of exclusion booleans.
            For example, a cell that was previously excluded with a
            a boolean mask (value of 1) may instead be converted to an
            inclusion weight value of 0.75, meaning that 75% of the area
            corresponding to that point should be included (i.e. the
            exclusion feature only intersected a small portion - 25% -
            of the cell). This percentage inclusion value is calculated
            by upscaling the output array using this input value,
            rasterizing the exclusion features onto it, and counting the
            number of resulting sub-cells excluded by the feature. For
            example, setting the value to `3` would split each output
            cell into nine sub-cells - 3 divisions in each dimension.
            After the feature is rasterized on this high-resolution
            sub-grid, the area of the non-excluded sub-cells is totaled
            and divided by the area of the original cell to obtain the
            final inclusion percentage. Therefore, a larger upscale
            factor results in more accurate percentage values. If `None`
            (or a value <= 1), this process is skipped and the output is
            a boolean exclusion mask. By default `None`.
        """
        self._rasterizer = Rasterizer(excl_fpath,
                                      weights_calculation_upscale_factor, hsds)
        super().__init__(excl_fpath, regulations, features, hsds)
        self._features_meta = GPKGMeta(self._features)

    def __repr__(self):
        msg = "{} for {}".format(self.__class__.__name__, self._excl_fpath)
        return msg

    @property
    def description(self):
        """str: Description to be added to excl H5."""
        return ('{} computed with a base setback distance of {} and a '
                'multiplier of {} for a total generic setback value of {} '
                '(local exclusions may differ).'
                .format(self.__class__,
                        self._regulations.base_setback_dist,
                        self._regulations.multiplier,
                        self._regulations.generic))

    @property
    def no_exclusions_array(self):
        """np.array: Array representing no exclusions. """
        return self._rasterizer.rasterize(shapes=None)

    @property
    def exclusion_merge_func(self):
        """callable: Function to merge overlapping exclusion layers. """
        return np.minimum if self._rasterizer.inclusions else np.maximum

    def pre_process_regulations(self):
        """Reduce regulations to state corresponding to features."""
        feats_crs = self._features_meta.crs
        xmin, ymin, xmax, ymax = self._features_meta.bbox
        regulations_df = self.regulations_table.to_crs(feats_crs)
        regulations_df = regulations_df.cx[xmin:xmax, ymin:ymax]
        regulations_df = regulations_df.to_crs(crs=self.profile['crs'])
        self._regulations.df = regulations_df.reset_index(drop=True)

        mask = self._regulation_table_mask()
        if not mask.any():
            msg = "Found no local regulations!"
            logger.warning(msg)
            warn(msg)

        self._regulations.df = (self.regulations_table[mask]
                                .reset_index(drop=True))
        logger.debug('Loaded and processed setback regulations for {} counties'
                     .format(len(self.regulations_table)))

    def _local_exclusions_arguments(self, regulation_value, county):
        """Compile and return arguments to `compute_local_exclusions`. """
        logger.debug("Selecting county IDs using bounds {}"
                     .format(county.total_bounds))
        county = (county.buffer(regulation_value * 1.1)
                  .to_crs(self._features_meta.crs))
        ids = self._features_meta.feat_ids_for_bbox(county.total_bounds)
        logger.debug("Calculating setbacks for counties with IDs {}"
                     .format(ids))

        for start in range(0, len(ids), self.NUM_FEATURES_PER_WORKER):
            end = start + self.NUM_FEATURES_PER_WORKER
            yield (ids[start:end], self._features,
                   self._features_meta.primary_key_column, self.profile['crs'],
                   self.FEATURE_FILTER_TYPE, self.BUFFER_TYPE,
                   self._rasterizer)

    @staticmethod
    def compute_local_exclusions(regulation_value, county, *args):
        """Compute local features setbacks.

        This method will compute the setbacks using a county-specific
        regulations file that specifies either a static setback or a
        multiplier value that will be used along with the base setback
        distance to compute the setback.

        Parameters
        ----------
        regulation_value : float | int
            Setback distance in meters.
        county : geopandas.GeoDataFrame
            Regulations for a single county.
        features_ids : iterable of ints
            List of tuple (or other iterable) of integer values
            corresponding to the ID of the features in the GeoPackage
            to load and compute exclusions for. Note that these ID
            values are the internal SQL table ID's stored with the
            features, NOT the index of the features when loaded using
            :func:`geopandas.read_file`.
        features_fp : path-like
            Path to the GeoPackage file containing the features to be
            loaded and used for the exclusion calculation.
        col : str
            Namer of the primary key column in the main SQL table of the
            GeoPackage. This should be the name of the column under
            which the `features_ids` can be found.
        crs : str
            String representation of the Coordinate Reference System of
            the output exclusions array.
        features_filter_type : str
            Key from the :attr:`FEATURE_FILTERS` dictionary that points
            to the feature filter function to use. This feature filter
            function filters the loaded features such that they are
            localized to the county bounds.
        buffer_type : str
            Key from the :attr:`BUFFERS` dictionary that points to the
            feature buffer function to use.
        rasterizer : Rasterizer
            Instance of `Rasterizer` class used to rasterize the
            buffered county features.

        Returns
        -------
        setbacks : ndarray
            Raster array of setbacks
        slices : 2-tuple of `slice`
            X and Y slice objects defining where in the original array
            the exclusion data should go.
        """
        (features_ids, features_fp, col, crs, features_filter_type,
         buffer_type, rasterizer) = args
        logger.debug('- Computing setbacks for county FIPS {}'
                     .format(county.iloc[0]['FIPS']))
        log_mem(logger)
        features = _load_features(features_ids, features_fp, col, crs)
        feature_bounds = _buffered_feature_bounds(features, rasterizer,
                                                  regulation_value)

        features = FEATURE_FILTERS[features_filter_type](features, county)
        features = BUFFERS[buffer_type](features, regulation_value)

        return _rasterize_within_window(features, feature_bounds, rasterizer)

    def compute_generic_exclusions(self, max_workers=None):
        """Compute generic setbacks.

        This method will compute the setbacks using a generic setback
        of `base_setback_dist * multiplier`.

        Parameters
        ----------
        max_workers : int, optional
            Number of workers to use for exclusions computation, if 1
            run in serial, if > 1 run in parallel with that many
            workers, if `None` run in parallel on all available cores.
            By default `None`.

        Returns
        -------
        setbacks : ndarray
            Raster array of setbacks
        """
        generic_regulations_dne = (self._regulations.generic is None
                                   or np.isclose(self._regulations.generic, 0))
        if generic_regulations_dne:
            return self.no_exclusions_array

        max_workers = max_workers or os.cpu_count()
        ids = self._features_meta.feat_ids
        num_feats = self._features_meta.num_feats
        pk = self._features_meta.primary_key_column
        crs = self.profile['crs']
        exclusions = None
        if max_workers > 1:
            msg = ("Computing generic setbacks from {:,} features using {} "
                   "workers".format(num_feats, max_workers))
            logger.debug(msg)

            loggers = [__name__, 'reVX', 'rex']
            spp_kwargs = {"max_workers": max_workers, "loggers": loggers}
            with SpawnProcessPool(**spp_kwargs) as exe:
                exclusions = self._compute_generic_exclusions_in_chunks(
                    exe, max_workers, ids, pk, crs)
        else:
            logger.info("Computing generic setbacks from {} features in "
                        "serial.".format(num_feats))
            for start in range(0, len(ids), self.NUM_FEATURES_PER_WORKER):
                end = start + self.NUM_FEATURES_PER_WORKER
                out = _compute_exclusions(ids[start:end], self._features, pk,
                                          crs, self.BUFFER_TYPE,
                                          self._regulations.generic,
                                          self._rasterizer)
                new_exclusions, slices = out
                exclusions = self._combine_exclusions(exclusions,
                                                      new_exclusions,
                                                      slices=slices)
                msg = ("Computed generic setbacks for {:,}/{:,} features"
                       .format(end, num_feats))
                logger.info(msg)

        if exclusions is None:
            exclusions = self.no_exclusions_array

        return exclusions

    def _compute_generic_exclusions_in_chunks(self, exe, max_submissions,
                                              ids, pk, crs):
        """Compute exclusions in parallel using futures. """
        futures, exclusions = {}, None

        futures = []
        start_inds = range(0, len(ids), self.NUM_FEATURES_PER_WORKER)
        for ind, start in enumerate(start_inds, start=1):
            end = start + self.NUM_FEATURES_PER_WORKER
            future = exe.submit(_compute_exclusions, ids[start:end],
                                self._features, pk, crs,
                                self.BUFFER_TYPE,
                                self._regulations.generic,
                                self._rasterizer)
            futures.append(future)
            if ind % max_submissions == 0:
                exclusions = self._collect_ge_futures(futures, exclusions)
                msg = ("Computed generic setbacks for {:,}/{:,} features"
                       .format(end, len(ids)))
                logger.info(msg)

        exclusions = self._collect_ge_futures(futures, exclusions)
        return exclusions

    def _collect_ge_futures(self, futures, exclusions):
        """Collect all futures from the input dictionary. """
        logger.debug(f"Collecting {len(futures)} futures...")
        log_mem(logger)
        for future in as_completed(futures):
            new_exclusions, slices = future.result()
            exclusions = self._combine_exclusions(exclusions,
                                                  new_exclusions,
                                                  slices=slices)
        futures.clear()
        logger.debug("Finished collecting futures chunk!")
        log_mem(logger)
        return exclusions

    def _regulation_table_mask(self):
        """Return the regulation table mask for setback feature. """
        features = (self.regulations_table['Feature Type']
                    .isin(self.FEATURE_TYPES))
        not_excluded = ~(self.regulations_table['Feature Subtype']
                         .isin(self.FEATURE_SUBTYPES_TO_EXCLUDE))
        return features & not_excluded

    @property
    @abstractmethod
    def FEATURE_TYPES(self):
        """set: Feature type names using in the regulations file. """
        raise NotImplementedError

    @property
    @abstractmethod
    def FEATURE_SUBTYPES_TO_EXCLUDE(self):
        """set: Feature subtype names to exclude from regulations file. """
        raise NotImplementedError

    @property
    @abstractmethod
    def BUFFER_TYPE(self):
        """str: Key in `BUFFERS` pointing to buffer to use. """
        raise NotImplementedError

    @property
    @abstractmethod
    def FEATURE_FILTER_TYPE(self):
        """str: Key in `FEATURE_FILTERS` pointing to feature filter to use. """
        raise NotImplementedError

    @property
    @abstractmethod
    def NUM_FEATURES_PER_WORKER(self):
        """int: Number of features each worker processes at one time. """
        raise NotImplementedError


def _parse_excl_properties(excl_fpath, hsds=False):
    """Parse shape, chunk size, and profile from exclusions file.

    Parameters
    ----------
    excl_fpath : str
        Path to .h5 file containing exclusion layers, will also be
        the location of any new setback layers
    hsds : bool, optional
        Boolean flag to use h5pyd to handle .h5 'files' hosted on
        AWS behind HSDS. By default `False`.

    Returns
    -------
    shape : tuple
        Shape of exclusions datasets
    profile : str
        GeoTiff profile for exclusions datasets
    """
    with ExclusionLayers(excl_fpath, hsds=hsds) as exc:
        dset_shape = exc.shape
        profile = exc.profile

    if len(dset_shape) < 3:
        dset_shape = (1, ) + dset_shape

    logger.debug('Exclusions properties:\n'
                 'shape : {}\n'
                 'profile : {}\n'
                 .format(dset_shape, profile))

    return dset_shape, profile


def _load_features(features_ids, features_fp, col, crs):
    """Load the `features_ids` from the `features_fp`. """
    ids = ",".join(map(str, features_ids))
    logger.debug("  Loading {} features from {}".format(len(features_ids),
                                                        features_fp))
    features = gpd.read_file(features_fp,
                             where="{} in ({})".format(col, ids),
                             engine="pyogrio")
    features = features.to_crs(crs=crs)
    features["geometry"] = features.apply(_make_row_shape_valid, axis=1)

    logger.debug("Loaded {} features".format(len(features)))
    logger.debug("Features total bounds: {}".format(features.total_bounds))
    log_mem(logger)

    return features


def _make_row_shape_valid(row):
    """Make a row shape valid using shapely `make_valid`"""
    return unary_union(make_valid(row["geometry"]))


def _compute_exclusions(features_ids, features_fp, col, crs, buffer_type,
                        setback, rasterizer):
    """Compute exclusions by loading features, buffering, and rasterizing. """
    setbacks, feature_bounds = _load_and_buffer(features_ids, features_fp,
                                                col, crs, buffer_type,
                                                setback, rasterizer)
    if setbacks is None:
        return None, None

    return _rasterize_within_window(setbacks, feature_bounds, rasterizer)


def _load_and_buffer(features_ids, features_fp, col, crs, buffer_type,
                     setback, rasterizer):
    """Load features and immediately buffer them.

    The intention is to keep loading and buffering in one function so
    that large sets of features get dropped immediately instead of
    hanging in memory during rasterization.
    """
    features = _load_features(features_ids, features_fp, col, crs)
    excl_array_bbox = transform.array_bounds(*rasterizer.arr_shape[1:],
                                             rasterizer.transform)
    if coords.disjoint_bounds(excl_array_bbox, features.total_bounds):
        return None, None

    logger.debug(f"Buffering {len(features)} features...")
    setbacks = BUFFERS[buffer_type](features, setback)
    feature_bounds = _buffered_feature_bounds(features, rasterizer, setback)
    return setbacks, feature_bounds


def _buffered_feature_bounds(features, rasterizer, regulation_value):
    """Calculate the buffered feature bounds"""
    buffer_len = max(abs(rasterizer.transform.a), abs(rasterizer.transform.e))
    bound_buffer = regulation_value * 2 + buffer_len
    return features.buffer(bound_buffer).total_bounds


def _rasterize_within_window(features, bounds, rasterizer):
    """Rasterize the features using the geoseries bounding box as a window. """
    window = _cropped_window(bounds, rasterizer.transform,
                             rasterizer.arr_shape[1:])
    logger.debug(f"Rasterizing {len(features)} features using {window!r}")
    exclusions = rasterizer.rasterize(features, window=window)
    log_mem(logger)
    logger.debug(f"Exclusion mem size: {exclusions.nbytes / 1048576:.2f}MB")
    return exclusions, window.toslices()


def _cropped_window(bounds, raster_transform, shape):
    """Calculate the raster array window corresponding to the bounding box."""
    left, bottom, right, top = bounds

    rows, cols = transform.rowcol(raster_transform,
                                  [left, right, right, left],
                                  [top, top, bottom, bottom],
                                  op=float)

    row_start = max(floor(min(rows)), 0)
    col_start = max(floor(min(cols)), 0)
    row_stop = min(ceil(max(rows)), shape[0])
    col_stop = min(ceil(max(cols)), shape[1])
    return windows.Window(col_off=col_start, row_off=row_start,
                          width=max(col_stop - col_start, 0),
                          height=max(row_stop - row_start, 0))




# -*- coding: utf-8 -*-
"""
reVX ReEDS column addition utilities
"""
import os
import json
import pandas as pd
from warnings import warn

from rex import Resource
from reVX.utilities.region_classifier import RegionClassifier
from reVX.utilities.utilities import load_fips_to_state_map
from reVX.version import __version__


UTILITY_DIR = os.path.dirname(os.path.realpath(__file__))
CONFIG_DIR = os.path.join(UTILITY_DIR, "config")


def add_county_info(data_frame, regions):
    """Add county info to a Pandas DataFrame with coordinates.

    The input DataFrame must have latitude and longitude columns.

    Parameters
    ----------
    data_frame : pandas.DataFrame
        A pandas data frame with latitude and longitude coordinates.
    regions : str | GeoDataFrame
        Path to regions shapefile containing labeled geometries or
        a pre-loaded GeoDataFrame.

    Returns
    -------
    pandas.DataFrame
        A pandas data frame with all initial input data plus three new
        columns: "cnty_fips", "state", and "county". "cnty_fips" is a
        five-digit county code, while "state" and "county" are the state
        and county names, respectively.
    """
    data_frame = data_frame.drop(columns=["cnty_fips", "county"],
                                 errors="ignore")
    data_frame = _classify(data_frame, "GEOID", regions)
    data_frame = _classify(data_frame, "NAME", regions)
    data_frame = data_frame.rename(columns={"GEOID": "cnty_fips",
                                            "NAME": "county"})

    cmap = load_fips_to_state_map()
    data_frame["state"] = data_frame["cnty_fips"].apply(
        lambda code: cmap[code[:2]])
    return data_frame


def _classify(data_frame, col, regions):
    """Classify a single county column for the input DataFrame"""
    classifier = RegionClassifier(data_frame, regions, col)
    data_frame = classifier.classify(force=True)
    return data_frame.drop(columns="geometry", errors="ignore")


def _lowercase_alpha_only(in_str):
    """Convert a string to lowercase alphabetic values only (a-z)"""
    return ''.join(filter(str.isalpha, in_str.casefold()))


def add_nrel_regions(data_frame):
    """Add NREL Regions info to a Pandas DataFrame with coordinates.

    The input DataFrame must have a "state" column containing teh state
    name for each row.

    Parameters
    ----------
    data_frame : pandas.DataFrame
        A pandas data frame with "state" column.

    Returns
    -------
    pandas.DataFrame
        A pandas data frame with an extra "nrel_region" column.
    """
    if "state" not in data_frame:
        raise KeyError("Input DataFrame missing required column 'state'")

    with open(os.path.join(CONFIG_DIR, "nrel_regions.json")) as fh:
        nrel_regions = json.load(fh)

    regions = {_lowercase_alpha_only(key): val
               for key, val in nrel_regions.items()}

    states = data_frame["state"].apply(_lowercase_alpha_only)
    data_frame["nrel_region"] = states.map(regions)
    return data_frame


def add_extra_data(data_frame, extra_data, merge_col="sc_point_gid"):
    """Add extra data to a Pandas DataFrame from a list of input files.

    Parameters
    ----------
    data_frame : pandas.DataFrame
        A pandas data frame with initial data. Must have `merge_col`
        column if extracting data from HDF5 files.
    extra_data : list of dicts
        A list of dictionaries, where each dictionary contains two keys.
        The first key is "source", and its value must either be a
        dictionary of `field: value` pairs or a path to the extra data
        being extracted. The latter must be a path pointing to an
        HDF5 or JSON file (i.e. it must end in ".h5" or ".json"). The
        second key is "dsets", and it points to a list of dataset names
        to extract from `source`. For JSON and dictionary data
        extraction, the values of the datasets must either be scalars or
        must match the length of the input `data_frame`. For HDF5 data,
        the datasets must be 1D datasets, and they will be merged with
        the input `data_frame` on `merge_col` (column must be in the
        HDF5 file meta). By default, ``None``.
    merge_col : str, optional
        Name of column used to merge the data in the input
        `data_frame` with the data in the HDF5 file. Note that this
        column must be present in both the `data_frame` as well as the
        HDF5 file meta.

    Returns
    -------
    pandas.DataFrame
        A pandas data frame with extra data added from input files.
    """
    for data_info in extra_data:
        source = data_info["source"]
        dsets = data_info["dsets"]

        if isinstance(source, dict):
            extra_data = source
        elif str(source).endswith(".json"):
            with open(source, "r") as fh:
                extra_data = json.load(fh)
        elif str(source).endswith(".h5"):
            with Resource(source) as res:
                extra_data = res.meta[[merge_col]].copy()
                for dset in dsets:
                    extra_data[dset] = res[dset]
            extra_data = pd.merge(data_frame[[merge_col]], extra_data,
                                  on=merge_col)
            extra_data = {dset: extra_data[dset].values
                          for dset in dsets}
        else:
            msg = ("File format not currently supported for file: {}"
                   .format(source))
            warn(msg)
            continue

        for dset in dsets:
            data_frame[dset] = extra_data[dset]

    return data_frame


def add_reeds_columns(supply_curve_fpath, out_fp=None, capacity_col="capacity",
                      extra_data=None, merge_col="sc_point_gid",
                      filter_out_zero_capacity=True, rename_mapping=None,
                      regions=None):
    """Add columns to supply curve required by ReEDS.

    This method will add columns like "cnty_fips", "state", "county",
    "nrel_region", "eos_mult", and "reg_mult". This method also allows
    you to add extra columns from H5 or JSON files.

    Parameters
    ----------
    supply_curve_fpath : str
        Path to input supply curve. Should have standard reV supply
        curve output columns (e.g. latitude, longitude, capacity,
        sc_point_gid, etc.). If running from CLI, this can be a list
        of supply curve paths.
    out_fp : str, optional
        Path to output file for supply curve with new columns. If
        ``None``, the supply curve will be overwritten (i.e. the data
        will be written to `supply_curve_fpath`). If running from CLI,
        this can be a list output paths (length *must* match length of
        `supply_curve_fpath`). By default, ``None``.
    capacity_col : str, optional
        Name of capacity column. This is used to filter out sites with
        zero capacity, if that option is selected.
        By default, ``"capacity"``.
    extra_data : list of dicts, optional
        A list of dictionaries, where each dictionary contains two keys.
        The first key is "source", and its value must either be a
        dictionary of `field: value` pairs or a path to the extra data
        being extracted. The latter must be a path pointing to an
        HDF5 or JSON file (i.e. it must end in ".h5" or ".json"). The
        second key is "dsets", and it points to a list of dataset names
        to extract from `source`. For JSON and dictionary data
        extraction, the values of the datasets must either be scalars or
        must match the length of the input `data_frame`. For HDF5 data,
        the datasets must be 1D datasets, and they will be merged with
        the input `data_frame` on `merge_col` (column must be in the
        HDF5 file meta). By default, ``None``.
    merge_col : str, optional
        Name of column used to merge the data in the input supply curve
        with the data in the HDF5 file if `extra_data` is specified.
        Note that this column must be present in both the input supply
        curve as well as the HDF5 file meta.
        By default, ``"sc_point_gid"``.
    filter_out_zero_capacity : bool, optional
        Flag to filter out sites with zero capacity.
        By default, ``True``.
    rename_mapping : dict, optional
        Optional mapping of old column names to new column names. This
        mapping will be used to rename the columns in the supply curve
        towards the end of the procedure (after all extra columns except
        ``eos_mult`` and ``reg_mult`` have been added).
        By default, ``None`` (no renaming).
    regions : str, optional
        Path to a regions shapefile containing county geometries labeled
        with county FIPS values. You can download the data from
        `www2.census.gov/geo/tiger/TIGER2021/COUNTY/tl_2021_us_county.zip`.
        If ``None``, then no county info is added. By default, ``None``.

    Returns
    -------
    out_fpath : str
        Path to output file.
    """

    sc = pd.read_csv(supply_curve_fpath)
    if regions is not None:
        sc = add_county_info(sc, regions)
        sc = add_nrel_regions(sc)

    if extra_data:
        sc = add_extra_data(sc, extra_data, merge_col=merge_col)

    if filter_out_zero_capacity and capacity_col in sc:
        sc = sc[sc[capacity_col] > 0]

    rename_mapping = rename_mapping or {}
    sc = sc.rename(columns=rename_mapping)

    for col in ["eos_mult", "reg_mult"]:
        if col not in sc:
            sc[col] = 1

    sc = sc.reset_index(drop=True)

    out_fp = out_fp or supply_curve_fpath
    sc.to_csv(out_fp, index=False)
    return out_fp



# -*- coding: utf-8 -*-
"""
Module to compute least cost transmission paths, distances, AND costs
for one or more SC points
"""
from warnings import warn
import geopandas as gpd
import json
import logging
import numpy as np
import os
import time
import pandas as pd
import rasterio

from pyproj.crs import CRS
from scipy.spatial import cKDTree
from shapely.geometry import Point
from concurrent.futures import as_completed

from reV.handlers.exclusions import ExclusionLayers
from reV.supply_curve.extent import SupplyCurveExtent
from rex.utilities.execution import SpawnProcessPool
from rex.utilities.loggers import log_mem

from reVX.least_cost_xmission.config.constants import (CELL_SIZE,
                                                       TRANS_LINE_CAT,
                                                       LOAD_CENTER_CAT,
                                                       SINK_CAT,
                                                       SUBSTATION_CAT,
                                                       CLIP_RASTER_BUFFER,
                                                       NUM_NN_SINKS,
                                                       RESOLUTION,
                                                       MINIMUM_SPUR_DIST_KM,
                                                       BARRIERS_MULT,
                                                       BARRIER_H5_LAYER_NAME,
                                                       ISO_H5_LAYER_NAME)
from reVX.least_cost_xmission.least_cost_paths import LeastCostPaths
from reVX.least_cost_xmission.trans_cap_costs import (TransCapCosts,
                                                      RegionalTransCapCosts)

logger = logging.getLogger(__name__)


class LeastCostXmission(LeastCostPaths):
    """
    Compute Least Cost tie-line paths and full transmission cap cost
    for all possible connections to all supply curve points
    """
    _TCC_CLASS = TransCapCosts

    def __init__(self, cost_fpath, features_fpath, resolution=RESOLUTION,
                 xmission_config=None, min_line_length=MINIMUM_SPUR_DIST_KM,
                 tb_layer_name=BARRIER_H5_LAYER_NAME,
                 iso_regions_layer_name=ISO_H5_LAYER_NAME):
        """
        Parameters
        ----------
        cost_fpath : str
            Path to h5 file with cost rasters and other required layers
        features_fpath : str
            Path to GeoPackage with transmission features
        resolution : int, optional
            SC point resolution, by default 128
        xmission_config : str | dict | XmissionConfig, optional
            Path to Xmission config .json, dictionary of Xmission config
            .jsons, or preloaded XmissionConfig objects, by default None
        min_line_length : int | float, optional
            Minimum line length in km, by default 0.
        tb_layer_name : str, default=:obj:`BARRIER_H5_LAYER_NAME`
            Name of transmission barrier layer in `cost_fpath` file.
            This layer defines the multipliers applied to the cost layer
            to determine LCP routing (but does not actually affect
            output costs). By default, :obj:`BARRIER_H5_LAYER_NAME`.
        iso_regions_layer_name : str, default=:obj:`ISO_H5_LAYER_NAME`
            Name of ISO regions layer in `cost_fpath` file. The layer
            maps pixels to ISO region ID's (1, 2, 3, 4, etc.) .
            By default, :obj:`ISO_H5_LAYER_NAME`.
        """
        self._cost_fpath = cost_fpath
        self._tb_layer_name = tb_layer_name
        self._iso_layer_name = iso_regions_layer_name
        self._check_layers()

        self._config = self._TCC_CLASS._parse_config(
            xmission_config=xmission_config)

        (self._sc_points, self._features,
         self._sub_lines_mapping, self._shape) =\
            self._map_to_costs(cost_fpath, features_fpath,
                               resolution=resolution,
                               iso_layer_name=self._iso_layer_name)
        self._tree = None
        self._sink_coords = None
        self._min_line_len = min_line_length

        logger.debug('{} initialized'.format(self))

    def __repr__(self):
        msg = ("{} to be computed for {} sc_points and {} features"
               .format(self.__class__.__name__,
                       len(self.sc_points),
                       len(self.features)))

        return msg

    def _check_layers(self):
        """Check to make sure the required layers are in cost_fpath. """
        self._check_layers_in_h5(self._cost_fpath,
                                 [self._tb_layer_name, self._iso_layer_name])

    @property
    def sc_points(self):
        """
        Table of supply curve points

        Returns
        -------
        gpd.GeoDataFrame
        """
        return self._sc_points

    @property
    def features(self):
        """
        Table of features to compute paths for

        Returns
        -------
        pandas.DataFrame
        """
        return self._features

    @property
    def sub_lines_mapping(self):
        """
        Series mapping substations  to the transmission lines connected
        to each substation

        Returns
        -------
        pandas.Series
        """
        return self._sub_lines_mapping

    @property
    def sink_coords(self):
        """
        Inf sink coordinates (row, col)

        Returns
        -------
        ndarray
        """
        if self._sink_coords is None:
            mask = self.features['category'] == SINK_CAT
            self._sink_coords = self.features.loc[mask, ['row', 'col']].values

        return self._sink_coords

    @property
    def sink_tree(self):
        """
        cKDTree for infinite sinks

        Returns
        -------
        cKDTree
        """
        if self._tree is None:
            self._tree = cKDTree(self.sink_coords)

        return self._tree

    @staticmethod
    def _load_trans_feats(features_fpath):
        """
        Load existing transmission features from disk. Substations will
        be loaded from cache file if it exists

        Parameters
        ----------
        features_fpath : str
            Path to GeoPackage with trans features

        Returns
        -------
        features : gpd.GeoDataFrame
            DataFrame of transmission features
        sub_line_map : pandas.Series
            Mapping of sub-station trans_gid to connected transmission
            line trans_gids
        """
        logger.debug('Loading transmission features')
        features = gpd.read_file(features_fpath)
        features = features.drop(columns=['bgid', 'egid', 'cap_left'],
                                 errors='ignore')
        mapping = {'gid': 'trans_gid', 'trans_gids': 'trans_line_gids'}
        features = features.rename(columns=mapping)

        features['min_volts'] = 0
        features['max_volts'] = 0

        # Transmission lines
        mask = features['category'] == TRANS_LINE_CAT
        voltage = features.loc[mask, 'voltage'].values
        features.loc[mask, 'min_volts'] = voltage
        features.loc[mask, 'max_volts'] = voltage

        # Load Center and Sinks
        mask = features['category'].isin([LOAD_CENTER_CAT, SINK_CAT])
        features.loc[mask, 'min_volts'] = 1
        features.loc[mask, 'max_volts'] = 9999

        sub_lines_map = {}
        mask = features['category'] == SUBSTATION_CAT
        bad_subs = np.zeros(len(features), dtype=bool)
        for idx, row in features.loc[mask].iterrows():
            gid = row['trans_gid']
            lines = row['trans_line_gids']
            if isinstance(lines, str):
                lines = json.loads(lines)

            sub_lines_map[gid] = lines
            lines_mask = features['trans_gid'].isin(lines)
            voltage = features.loc[lines_mask, 'voltage'].values

            if np.max(voltage) >= 69:
                features.loc[idx, 'min_volts'] = np.min(voltage)
                features.loc[idx, 'max_volts'] = np.max(voltage)
            else:
                bad_subs[idx] = True

        if any(bad_subs):
            msg = ("The following sub-stations do not have the minimum "
                   "required voltage of 69 kV and will be dropped:\n{}"
                   .format(features.loc[bad_subs, 'trans_gid']))
            logger.warning(msg)
            warn(msg)
            features = features.loc[~bad_subs].reset_index(drop=True)

        return features, pd.Series(sub_lines_map)

    @staticmethod
    def _create_sc_points(cost_fpath, resolution=RESOLUTION):
        """
        Load SC points, covert row/col to array wide, and determine x/y
        for reV projection

        Parameters
        ----------
        cost_fpath : str
            Path to h5 file with cost rasters and other required layers
        resolution : int, optional
            SC point resolution, by default 128

        Returns
        sc_points : gpd.GeoDataFrame
            SC points
        """
        logger.debug('Loading Supply Curve Points for res: {}'
                     .format(resolution))
        sce = SupplyCurveExtent(cost_fpath, resolution=resolution)
        sc_points = sce.points.rename(columns={'row_ind': 'sc_row_ind',
                                               'col_ind': 'sc_col_ind'})
        shape = sce.excl_shape
        sc_points['sc_point_gid'] = sc_points.index.values

        row = np.round(sc_points['sc_row_ind'] * resolution + resolution / 2)
        row = np.where(row >= shape[0], shape[0] - 1, row)
        sc_points['row'] = row.astype(int)

        col = np.round(sc_points['sc_col_ind'] * resolution + resolution / 2)
        col = np.where(col >= shape[1], shape[1] - 1, col)
        sc_points['col'] = col.astype(int)

        return sc_points

    @staticmethod
    def _get_feature_cost_indices(features, crs, transform, shape):
        """
        Map features to cost row, col indices using rasterio transform

        Parameters
        ----------
        features : gpd.GeoDataFrame
            GeoDataFrame of features to map to cost raster
        crs : pyproj.crs.CRS
            CRS of cost raster
        transform : raster.Affine
            Transform of cost raster
        shape : tuple
            Cost raster shape

        Returns
        -------
        row : ndarray
            Vector of row indices for each feature
        col : ndarray
            Vector of col indices for each features
        mask : ndarray
            Boolean mask of features with indices outside of cost raster
        """
        row, col, mask = super(LeastCostXmission,
                               LeastCostXmission)._get_feature_cost_indices(
            features, crs, transform, shape)

        t_lines = features['category'] == TRANS_LINE_CAT
        mask |= t_lines

        row[t_lines] = np.where(row[t_lines] >= 0, row[t_lines], 0)
        row[t_lines] = np.where(row[t_lines] < shape[0], row[t_lines],
                                shape[0] - 1)
        col[t_lines] = np.where(col[t_lines] >= 0, col[t_lines], 0)
        col[t_lines] = np.where(col[t_lines] < shape[1], col[t_lines],
                                shape[1] - 1)

        return row, col, mask

    @classmethod
    def _map_to_costs(cls, cost_fpath, features_fpath, resolution=RESOLUTION,
                      iso_layer_name=ISO_H5_LAYER_NAME):
        """
        Map supply curve points and transmission features to cost array
        pixel indices

        Parameters
        ----------
        cost_fpath : str
            Path to h5 file with cost rasters and other required layers
        features_fpath : str
            Path to GeoPackage with transmission features
        resolution : int, optional
            SC point resolution, by default 128
        iso_layer_name : str, default=:obj:`ISO_H5_LAYER_NAME`
            Name of ISO regions layer in `cost_fpath` file. The layer
            maps pixels to ISO region ID's (1, 2, 3, 4, etc.) .
            By default, :obj:`ISO_H5_LAYER_NAME`.

        Returns
        -------
        sc_points : gpd.GeoDataFrame
            Table of supply curve points to connect to transmission
        features : gpd.GeoDataFrame
            Table of transmission features
        sub_lines_map : pandas.Series
            Series mapping substations  to the transmission lines
            connected to each substation
        shape: Tuple[int, int]
            Shape of h5 costs raster
        """
        with ExclusionLayers(cost_fpath) as f:
            crs = CRS.from_string(f.crs)
            transform = rasterio.Affine(*f.profile['transform'])
            shape = f.shape
            regions = f[iso_layer_name]

        features, sub_lines_map = cls._load_trans_feats(features_fpath)
        row, col, mask = cls._get_feature_cost_indices(features, crs,
                                                       transform, shape)
        if any(~mask):
            msg = ("The following features are outside of the cost exclusion "
                   "domain and will be dropped:\n{}"
                   .format(features.loc[~mask, 'trans_gid']))
            logger.warning(msg)
            warn(msg)
            row = row[mask]
            col = col[mask]
            features = features.loc[mask].reset_index(drop=True)

        features['row'] = row
        features['col'] = col
        features['region'] = regions[row, col]

        logger.debug('Converting SC points to GeoDataFrame')
        sc_points = cls._create_sc_points(cost_fpath, resolution=resolution)
        x, y = rasterio.transform.xy(transform, sc_points['row'].values,
                                     sc_points['col'].values)
        geo = [Point(xy) for xy in zip(x, y)]
        sc_points = gpd.GeoDataFrame(sc_points, crs=features.crs,
                                     geometry=geo)

        return sc_points, features, sub_lines_map, shape

    def _clip_to_sc_point(self, sc_point, tie_line_voltage,
                          nn_sinks=NUM_NN_SINKS,
                          clipping_buffer=CLIP_RASTER_BUFFER, radius=None,
                          expand_radius=True):
        """
        Clip costs raster to AOI around SC point, and get substations,
        load centers, and sinks within the clipped region.

        Parameters
        ----------
        sc_point : gpd.GeoSeries
            SC point to clip raster around
        nn_sinks : int, optional
            Number of nearest neighbor sinks to clip to
        clipping_buffer : float, optional
            Buffer to increase clipping radius by, by default 1.05
        radius : None | int, optional
            Force clipping radius. Transmission features beyond this
            radius will not be considered for connection with supply
            curve point. If ``None``, no radius is forced, and
            connections to all available transmission features are
            computed. By default, ``None``.
        expand_radius : bool, optional
            Option to expand radius to include at least one connection
            feature. Has no effect if ``radius=None``.
            By default, ``True``.

        Returns
        -------
        radius : int
            Clipping radius in cost raster pixels
        x_feats : pd.DataFrame
            Substations, load centers, sinks, and nearest points on
            t-lines to SC point
        """
        logger.debug('Clipping features to sc_point {}'.format(sc_point.name))

        sc_features = self.features.copy(deep=True)
        if len(self.sink_coords) > 2 or radius:
            row, col = sc_point[['row', 'col']].values

            if radius is None:
                _, pos = self.sink_tree.query([row, col], k=nn_sinks)
                radius = np.abs(self.sink_coords[pos] - np.array([row, col])
                                ).max()
                radius = int(np.ceil(radius * clipping_buffer))
                logger.debug('Radius to {} nearest sink is: {}'
                             .format(nn_sinks, radius))
            else:
                logger.debug('Using forced radius of {}'.format(radius))

            with ExclusionLayers(self._cost_fpath) as f:
                resolution = f.profile["transform"][0]

            sc_features = self._clip_to_radius(sc_point, radius, sc_features,
                                               clipping_buffer, resolution,
                                               expand_radius)

        mask = self.features['max_volts'] >= tie_line_voltage
        sc_features = sc_features.loc[mask].copy(deep=True)

        if sc_features.empty:
            return sc_features, None

        dists = (sc_features[['row', 'col']] - sc_point[['row', 'col']])
        radius = int(np.ceil(dists.abs().values.max() * clipping_buffer))
        logger.debug('{} transmission features found in clipped area with '
                     'radius {} and minimum max voltage of {}'
                     .format(len(sc_features), radius, tie_line_voltage))

        # Find t-lines connected to substations within clip
        logger.debug('Collecting transmission lines connected to substations')
        mask = sc_features['category'] == SUBSTATION_CAT
        if mask.any():
            trans_gids = sc_features.loc[mask, 'trans_gid'].values
            trans_gids = \
                np.concatenate(self.sub_lines_mapping.loc[trans_gids].values)
            trans_gids = np.unique(trans_gids)
            line_mask = self.features['trans_gid'].isin(trans_gids)
            trans_lines = self.features.loc[line_mask].copy(deep=True)
            line_mask = trans_lines['trans_gid'].isin(sc_features['trans_gid'])
            trans_lines = trans_lines.loc[~line_mask]
            logger.debug('Adding all {} transmission lines connected to '
                         'substations with minimum max voltage of {}'
                         .format(len(trans_lines), tie_line_voltage))
            sc_features = pd.concat([sc_features, trans_lines])

        return sc_features, radius

    def _clip_to_radius(self, sc_point, radius, sc_features, clipping_buffer,
                        resolution, expand_radius=True):
        """Clip features to radius.

        If no features are found within the initial radius, it is
        expanded (linearly by incrementally increasing the clipping
        buffer) until at least one connection feature is found.
        """
        if radius is None or len(sc_features) == 0:
            return sc_features

        radius_m = radius * resolution
        logger.debug('Clipping features to radius {}m'.format(radius_m))
        buffer = sc_point["geometry"].buffer(radius_m)
        clipped_sc_features = sc_features.clip(buffer)

        clipping_buffer -= 0.05
        while expand_radius and len(clipped_sc_features) <= 0:
            clipping_buffer += 0.05
            radius_m = radius * resolution * clipping_buffer
            buffer = sc_point["geometry"].buffer(radius_m)
            clipped_sc_features = sc_features.clip(buffer)

        logger.info('{} transmission features found in clipped area with '
                    'radius {}'
                    .format(len(clipped_sc_features), radius_m))
        return clipped_sc_features.copy(deep=True)

    def process_sc_points(self, capacity_class, cost_layers,
                          sc_point_gids=None, nn_sinks=NUM_NN_SINKS,
                          clipping_buffer=CLIP_RASTER_BUFFER,
                          barrier_mult=BARRIERS_MULT,
                          max_workers=None, save_paths=False, radius=None,
                          expand_radius=True, mp_delay=3, simplify_geo=None,
                          length_invariant_cost_layers=None,
                          tracked_layers=None, length_mult_kind="linear",
                          cell_size=CELL_SIZE):
        """
        Compute Least Cost Transmission for desired sc_points

        Parameters
        ----------
        capacity_class : str | int
            Capacity class of transmission features to connect supply
            curve points to
        cost_layers : List[str]
            List of layers in H5 that are summed to determine total
            costs raster used for routing. Costs and distances for each
            individual layer are also reported (e.g. wet and dry costs).
            Layer names may have curly brackets (``{}``), which will be
            filled in based on the capacity class input (e.g.
            "tie_line_costs_{}MW").
        sc_point_gids : list, optional
            List of sc_point_gids to connect to, by default connect to
            all
        nn_sinks : int, optional
            Number of nearest neighbor sinks to use for clipping radius
            calculation, by default 2
        clipping_buffer : float, optional
            Buffer to expand clipping radius by, by default 1.05
        barrier_mult : int, optional
            Transmission barrier multiplier, used when computing the
            least cost tie-line path, by default 100
        max_workers : int, optional
            Number of workers to use for processing, if 1 run in serial,
            if None use all available cores, by default None
        save_paths : bool, optional
            Flag to return least cost paths as a multi-line geometry,
            by default False
        radius : None | int, optional
            Force clipping radius. Trasmission features beyond this
            radius will not be considered for connection with supply
            curve point. If ``None``, no radius is forced, and
            connections to all available  transmission features are
            computed. By default, ``None``.
        expand_radius : bool, optional
            Option to expand radius to include at least one connection
            feature. Has no effect if ``radius=None``.
            By default, ``True``.
        mp_delay : float, optional
            Delay in seconds between starting multi-process workers.
            Useful for reducing memory spike at working startup.
        simplify_geo : float | None, optional
            If float, simplify geometries using this value
        length_invariant_cost_layers : List[str] | None, optional
            List of layers in H5 to be added to the cost raster. The
            costs specified by these layers are not scaled with distance
            traversed across the cell (i.e. fixed one-time costs for
            crossing these cells).
        tracked_layers : dict, optional
            Dictionary mapping layer names to strings, where the strings
            are numpy methods that should be applied to the layer along
            the LCP. For example,
            ``tracked_layers={'layer_1': 'mean', 'layer_2': 'max}``
            would report the average of ``layer_1`` values along the
            least cost path and the max of ``layer_2`` values along the
            least cost path. Examples of numpy methods (non-exhaustive):

                - mean
                - max
                - min
                - mode
                - median
                - std

            By default, ``None``, which does not track any extra layers.
        length_mult_kind : {"step", "linear"}, default="linear"
            Type of length multiplier calcualtion. "step" computes
            length multipliers using a step function, while "linear"
            computes the length multiplier using a linear interpolation
            between 0 amd 10 mile spur-line lengths.
            By default, ``"linear"``.
        cell_size : int, optional
            Side length of each cell, in meters. Cells are assumed to be
            square. By default, :obj:`CELL_SIZE`.

        Returns
        -------
        least_costs : pandas.DataFrame | gpd.GeoDataFrame
            Least cost connections between all supply curve points and
            the transmission features with the given capacity class that
            are within "nn_sink" nearest infinite sinks
        """
        max_workers = os.cpu_count() if max_workers is None else max_workers

        if sc_point_gids is None:
            sc_point_gids = self.sc_points['sc_point_gid'].values

        tie_line_voltage = self._config.capacity_to_kv(capacity_class)

        # format any curly brackets in layer names given by user
        cc_str = self._config._parse_cap_class(capacity_class)
        cap = self._config['power_classes'][cc_str]
        cost_layers = [layer.format(cap) for layer in cost_layers]

        logger.debug('Using a barrier multiplier of %s', barrier_mult)

        if max_workers > 1:
            logger.info('Computing Least Cost Transmission for %d SC points '
                        'in parallel on %d workers',
                        len(sc_point_gids), max_workers)
            least_costs = self._process_multi_core(
                capacity_class,
                cost_layers,
                tie_line_voltage,
                sc_point_gids=sc_point_gids,
                nn_sinks=nn_sinks,
                clipping_buffer=clipping_buffer,
                barrier_mult=barrier_mult,
                save_paths=save_paths,
                radius=radius,
                expand_radius=expand_radius,
                mp_delay=mp_delay,
                simplify_geo=simplify_geo,
                max_workers=max_workers,
                length_invariant_cost_layers=length_invariant_cost_layers,
                tracked_layers=tracked_layers,
                length_mult_kind=length_mult_kind,
                cell_size=cell_size)
        else:
            logger.info('Computing Least Cost Transmission for {:,} SC points '
                        'in serial'.format(len(sc_point_gids)))
            least_costs = self._process_single_core(
                capacity_class,
                cost_layers,
                tie_line_voltage,
                sc_point_gids=sc_point_gids,
                nn_sinks=nn_sinks,
                clipping_buffer=clipping_buffer,
                barrier_mult=barrier_mult,
                save_paths=save_paths,
                radius=radius,
                expand_radius=expand_radius,
                simplify_geo=simplify_geo,
                length_invariant_cost_layers=length_invariant_cost_layers,
                tracked_layers=tracked_layers,
                length_mult_kind=length_mult_kind,
                cell_size=cell_size)

        if not least_costs:
            return pd.DataFrame(columns=['sc_point_gid'])

        least_costs = pd.concat(least_costs).sort_values(['sc_point_gid',
                                                          'trans_gid'])
        capacity_class = self._config._parse_cap_class(capacity_class)
        least_costs['max_cap'] = self._config['power_classes'][capacity_class]
        lcp_frac = (len(least_costs['sc_point_gid'].unique())
                    / len(sc_point_gids) * 100)
        logger.info('{:.4f}% of requested sc point gids were successfully '
                    'mapped to transmission features'.format(lcp_frac))

        return least_costs.reset_index(drop=True)

    def _process_multi_core(self, capacity_class, cost_layers,
                            tie_line_voltage, sc_point_gids,
                            nn_sinks=NUM_NN_SINKS,
                            clipping_buffer=CLIP_RASTER_BUFFER,
                            barrier_mult=BARRIERS_MULT, max_workers=2,
                            save_paths=False, radius=None, expand_radius=True,
                            mp_delay=3, simplify_geo=None,
                            length_invariant_cost_layers=None,
                            tracked_layers=None,
                            length_mult_kind="linear",
                            cell_size=CELL_SIZE):
        """
        Compute Least Cost Transmission for desired sc_points using
        multiple cores.

        Parameters
        ----------
        capacity_class : str | int
            Capacity class of transmission features to connect supply
            curve points to
        cost_layers : List[str]
            List of layers in H5 that are summed to determine total
            costs raster used for routing. Costs and distances for each
            individual layer are also reported (e.g. wet and dry costs).
            Layer names may have curly brackets (``{}``), which will be
            filled in based on the capacity class input (e.g.
            "tie_line_costs_{}MW").
        tie_line_voltage : int
            Tie-line voltage (kV)
        sc_point_gids : list | set
            List of sc_point_gids to connect to, by default connect to
            all
        nn_sinks : int, optional
            Number of nearest neighbor sinks to use for clipping radius
            calculation, by default 2
        clipping_buffer : float, optional
            Buffer to expand clipping radius by, by default 1.05
        barrier_mult : int, optional
            Transmission barrier multiplier, used when computing the
            least cost tie-line path, by default 100
        max_workers : int, optional
            Number of workers to use for processing
        save_paths : bool, optional
            Flag to return least cost paths as a multi-line geometry,
            by default False
        radius : None | int, optional
            Force clipping radius. Transmission features beyond this
            radius will not be considered for connection with supply
            curve point. If ``None``, no radius is forced, and
            connections to all available transmission features are
            computed. By default, ``None``.
        expand_radius : bool, optional
            Option to expand radius to include at least one connection
            feature. Has no effect if ``radius=None``.
            By default, ``True``.
        mp_delay : float, optional
            Delay in seconds between starting multi-process workers.
            Useful for reducing memory spike at working startup.
        simplify_geo : float | None, optional
            If float, simplify geometries using this value
        length_invariant_cost_layers : List[str] | None, optional
            List of layers in H5 to be added to the cost raster. The
            costs specified by these layers are not scaled with distance
            traversed across the cell (i.e. fixed one-time costs for
            crossing these cells).
        tracked_layers : dict, optional
            Dictionary mapping layer names to strings, where the strings
            are numpy methods that should be applied to the layer along
            the LCP. For example,
            ``tracked_layers={'layer_1': 'mean', 'layer_2': 'max}``
            would report the average of ``layer_1`` values along the
            least cost path and the max of ``layer_2`` values along the
            least cost path. Examples of numpy methods (non-exhaustive):

                - mean
                - max
                - min
                - mode
                - median
                - std

            By default, ``None``, which does not track any extra layers.
        length_mult_kind : {"step", "linear"}, default="linear"
            Type of length multiplier calcualtion. "step" computes
            length multipliers using a step function, while "linear"
            computes the length multiplier using a linear interpolation
            between 0 amd 10 mile spur-line lengths.
            By default, ``"linear"``.
        cell_size : int, optional
            Side length of each cell, in meters. Cells are assumed to be
            square. By default, :obj:`CELL_SIZE`.

        Returns
        -------
        least_costs : pandas.DataFrame | gpd.GeoDataFrame
            Least cost connections between all supply curve points and
            the transmission features with the given capacity class that
            are within "nn_sink" nearest infinite sinks
        """
        loggers = [__name__, 'reV', 'reVX']
        with SpawnProcessPool(max_workers=max_workers, loggers=loggers) as exe:
            least_costs = self._compute_paths_in_chunks(
                exe, max_workers, sc_point_gids, tie_line_voltage,
                nn_sinks, clipping_buffer, radius, expand_radius, mp_delay,
                capacity_class, cost_layers, barrier_mult, save_paths,
                simplify_geo, length_invariant_cost_layers, tracked_layers,
                length_mult_kind, cell_size)

        return least_costs

    def _compute_paths_in_chunks(self, exe, max_submissions, sc_point_gids,
                                 tie_line_voltage, nn_sinks, clipping_buffer,
                                 radius, expand_radius, mp_delay,
                                 capacity_class, cost_layers, barrier_mult,
                                 save_paths, simplify_geo, li_cost_layers,
                                 tracked_layers, length_mult_kind, cell_size):
        """Compute LCP's in parallel using futures. """
        futures, paths = {}, []

        mask = self.sc_points.sc_point_gid.isin(sc_point_gids)
        sc_points = self.sc_points[mask]
        if sc_points.empty:
            return paths

        costs = _starting_costs(self._cost_fpath, sc_points, cost_layers,
                                li_cost_layers, barrier_mult,
                                self._tb_layer_name)
        if (costs < 0).any():
            bad_points = sc_points[costs < 0]
            bad_sc_ids = list(bad_points["sc_point_gid"].values)
            msg = ("Could not connect SC points with invalid starting costs "
                   f"to transmission features: {bad_sc_ids}")
            logger.warning(msg)
            warn(msg)

        num_jobs = 1
        for __, sc_point in sc_points[costs >= 0].iterrows():
            gid = sc_point['sc_point_gid']
            if gid not in sc_point_gids:
                continue

            sc_features, sc_radius = self._clip_to_sc_point(
                sc_point, tie_line_voltage, nn_sinks=nn_sinks,
                clipping_buffer=clipping_buffer, radius=radius,
                expand_radius=expand_radius)
            if sc_features.empty:
                continue

            future = exe.submit(self._TCC_CLASS.run,
                                self._cost_fpath,
                                sc_point.copy(deep=True),
                                sc_features, capacity_class,
                                cost_layers,
                                radius=sc_radius,
                                xmission_config=self._config,
                                tb_layer_name=self._tb_layer_name,
                                barrier_mult=barrier_mult,
                                iso_regions_layer_name=self._iso_layer_name,
                                min_line_length=self._min_line_len,
                                save_paths=save_paths,
                                simplify_geo=simplify_geo,
                                length_invariant_cost_layers=li_cost_layers,
                                length_mult_kind=length_mult_kind,
                                tracked_layers=tracked_layers,
                                cell_size=cell_size)
            futures[future] = None
            num_jobs += 1
            if mp_delay > 0 and num_jobs <= max_submissions:
                time.sleep(mp_delay)
            logger.debug('Submitted {} futures'.format(num_jobs))
            log_mem(logger)
            if num_jobs % max_submissions == 0:
                paths = _collect_future_chunks(futures, paths)
        paths = _collect_future_chunks(futures, paths)
        return paths

    def _process_single_core(self, capacity_class, cost_layers,
                             tie_line_voltage, sc_point_gids,
                             nn_sinks=NUM_NN_SINKS,
                             clipping_buffer=CLIP_RASTER_BUFFER,
                             barrier_mult=BARRIERS_MULT, save_paths=False,
                             radius=None, expand_radius=True,
                             simplify_geo=None,
                             length_invariant_cost_layers=None,
                             tracked_layers=None, length_mult_kind="linear",
                             cell_size=CELL_SIZE):
        """
        Compute Least Cost Transmission for desired sc_points with a
        single core.

        Parameters
        ----------
        capacity_class : str | int
            Capacity class of transmission features to connect supply
            curve points to
        cost_layers : List[str]
            List of layers in H5 that are summed to determine total
            costs raster used for routing. Costs and distances for each
            individual layer are also reported (e.g. wet and dry costs).
            Layer names may have curly brackets (``{}``), which will be
            filled in based on the capacity class input (e.g.
            "tie_line_costs_{}MW").
        tie_line_voltage : int
            Tie-line voltage (kV)
        sc_point_gids : list | set
            List of sc_point_gids to connect to, by default connect to
            all
        nn_sinks : int, optional
            Number of nearest neighbor sinks to use for clipping radius
            calculation, by default 2
        clipping_buffer : float, optional
            Buffer to expand clipping radius by, by default 1.05
        barrier_mult : int, optional
            Transmission barrier multiplier, used when computing the
            least cost tie-line path, by default 100
        save_paths : bool, optional
            Flag to return least cost paths as a multi-line geometry,
            by default False
        radius : None | int, optional
            Force clipping radius. Transmission feature beyond this
            radius will not be considered for connection with supply
            curve point. If ``None``, no radius is forced, and
            connections to all available transmission features are
            computed. By default, ``None``.
        expand_radius : bool, optional
            Option to expand radius to include at least one connection
            feature. Has no effect if ``radius=None``.
            By default, ``True``.
        simplify_geo : float | None, optional
            If float, simplify geometries using this value
        length_invariant_cost_layers : List[str] | None, optional
            List of layers in H5 to be added to the cost raster. The
            costs specified by these layers are not scaled with distance
            traversed across the cell (i.e. fixed one-time costs for
            crossing these cells).
        tracked_layers : dict, optional
            Dictionary mapping layer names to strings, where the strings
            are numpy methods that should be applied to the layer along
            the LCP. For example,
            ``tracked_layers={'layer_1': 'mean', 'layer_2': 'max}``
            would report the average of ``layer_1`` values along the
            least cost path and the max of ``layer_2`` values along the
            least cost path. Examples of numpy methods (non-exhaustive):

                - mean
                - max
                - min
                - mode
                - median
                - std

            By default, ``None``, which does not track any extra layers.
        length_mult_kind : {"step", "linear"}, default="linear"
            Type of length multiplier calcualtion. "step" computes
            length multipliers using a step function, while "linear"
            computes the length multiplier using a linear interpolation
            between 0 amd 10 mile spur-line lengths.
            By default, ``"linear"``.
        cell_size : int, optional
            Side length of each cell, in meters. Cells are assumed to be
            square. By default, :obj:`CELL_SIZE`.

        Returns
        -------
        least_costs : pandas.DataFrame | gpd.GeoDataFrame
            Least cost connections between all supply curve points and
            the transmission features with the given capacity class that
            are within "nn_sink" nearest infinite sinks
        """
        least_costs = []
        count = 0
        for _, sc_point in self.sc_points.iterrows():
            gid = sc_point['sc_point_gid']
            if gid in sc_point_gids:
                sc_features, sc_radius = self._clip_to_sc_point(
                    sc_point, tie_line_voltage, nn_sinks=nn_sinks,
                    clipping_buffer=clipping_buffer, radius=radius,
                    expand_radius=expand_radius)
                if sc_features.empty:
                    continue
                sc_costs = self._TCC_CLASS.run(
                    self._cost_fpath,
                    sc_point.copy(deep=True),
                    sc_features, capacity_class,
                    radius=sc_radius,
                    xmission_config=self._config,
                    tb_layer_name=self._tb_layer_name,
                    barrier_mult=barrier_mult,
                    iso_regions_layer_name=self._iso_layer_name,
                    min_line_length=self._min_line_len,
                    save_paths=save_paths,
                    simplify_geo=simplify_geo,
                    cost_layers=cost_layers,
                    length_invariant_cost_layers=length_invariant_cost_layers,
                    length_mult_kind=length_mult_kind,
                    tracked_layers=tracked_layers,
                    cell_size=cell_size)

                if sc_costs is not None:
                    least_costs.append(sc_costs)

                count += 1
                logger.info('SC point {} of {} complete!'
                            .format(count, len(sc_point_gids)))
                log_mem(logger)
        return least_costs

    @classmethod
    def run(cls, cost_fpath, features_fpath, capacity_class, cost_layers,
            resolution=RESOLUTION, xmission_config=None,
            min_line_length=MINIMUM_SPUR_DIST_KM, sc_point_gids=None,
            nn_sinks=NUM_NN_SINKS, clipping_buffer=CLIP_RASTER_BUFFER,
            tb_layer_name=BARRIER_H5_LAYER_NAME, barrier_mult=BARRIERS_MULT,
            iso_regions_layer_name=ISO_H5_LAYER_NAME, max_workers=None,
            save_paths=False, radius=None, expand_radius=True, mp_delay=3,
            simplify_geo=None, length_invariant_cost_layers=None,
            tracked_layers=None, length_mult_kind="linear",
            cell_size=CELL_SIZE):
        """
        Find Least Cost Transmission connections between desired
        sc_points to given transmission features for desired capacity
        class

        Parameters
        ----------
        cost_fpath : str
            Path to h5 file with cost rasters and other required layers
        features_fpath : str
            Path to GeoPackage with transmission features
        capacity_class : str | int
            Capacity class of transmission features to connect supply
            curve points to
        cost_layers : List[str]
            List of layers in H5 that are summed to determine total
            costs raster used for routing. Costs and distances for each
            individual layer are also reported (e.g. wet and dry costs).
            Layer names may have curly brackets (``{}``), which will be
            filled in based on the capacity class input (e.g.
            "tie_line_costs_{}MW").
        resolution : int, optional
            SC point resolution, by default 128
        xmission_config : str | dict | XmissionConfig, optional
            Path to Xmission config .json, dictionary of Xmission config
            .jsons, or preloaded XmissionConfig objects, by default None
        min_line_length : int | float, optional
            Minimum line length in km, by default 0.
        sc_point_gids : list, optional
            List of sc_point_gids to connect to, by default None
        nn_sinks : int, optional
            Number of nearest neighbor sinks to use for clipping radius
            calculation, by default 2
        clipping_buffer : float, optional
            Buffer to expand clipping radius by, by default 1.05
        tb_layer_name : str, default=:obj:`BARRIER_H5_LAYER_NAME`
            Name of transmission barrier layer in `cost_fpath` file.
            This layer defines the multipliers applied to the cost layer
            to determine LCP routing (but does not actually affect
            output costs). By default, :obj:`BARRIER_H5_LAYER_NAME`.
        barrier_mult : int, optional
            Transmission barrier multiplier, used when computing the
            least cost tie-line path, by default 100
        iso_regions_layer_name : str, default=:obj:`ISO_H5_LAYER_NAME`
            Name of ISO regions layer in `cost_fpath` file. The layer
            maps pixels to ISO region ID's (1, 2, 3, 4, etc.) .
            By default, :obj:`ISO_H5_LAYER_NAME`.
            least cost tie-line path, by default 100
        max_workers : int, optional
            Number of workers to use for processing, if 1 run in serial,
            if None use all available cores, by default None
        save_paths : bool, optional
            Flag to return least costs path as a multi-line geometry,
            by default False
        radius : None | int, optional
            Force clipping radius. Transmission features beyond this
            radius wil not be considered for connection with supply
            curve point. If ``None``, no radius is forced, and
            connections to all available transmission features are
            computed. By default, ``None``.
        expand_radius : bool, optional
            Option to expand radius to include at least one connection
            feature. Has no effect if ``radius=None``.
            By default, ``True``.
        mp_delay : float, default=3.0
            Delay in seconds between starting multi-process workers.
            Useful for reducing memory spike at working startup.
        simplify_geo : float | None, optional
            If float, simplify geometries using this value
        length_invariant_cost_layers : List[str] | None, optional
            List of layers in H5 to be added to the cost raster. The
            costs specified by these layers are not scaled with distance
            traversed across the cell (i.e. fixed one-time costs for
            crossing these cells).
        tracked_layers : dict, optional
            Dictionary mapping layer names to strings, where the strings
            are numpy methods that should be applied to the layer along
            the LCP. For example,
            ``tracked_layers={'layer_1': 'mean', 'layer_2': 'max}``
            would report the average of ``layer_1`` values along the
            least cost path and the max of ``layer_2`` values along the
            least cost path. Examples of numpy methods (non-exhaustive):

                - mean
                - max
                - min
                - mode
                - median
                - std

            By default, ``None``, which does not track any extra layers.
        length_mult_kind : {"step", "linear"}, default="linear"
            Type of length multiplier calcualtion. "step" computes
            length multipliers using a step function, while "linear"
            computes the length multiplier using a linear interpolation
            between 0 amd 10 mile spur-line lengths.
            By default, ``"linear"``.
        cell_size : int, optional
            Side length of each cell, in meters. Cells are assumed to be
            square. By default, :obj:`CELL_SIZE`.

        Returns
        -------
        least_costs : pandas.DataFrame | gpd.DataFrame
            Least cost connections between all supply curve points and
            the transmission features with the given capacity class that
            are within "nn_sink" nearest infinite sinks
        """
        ts = time.time()
        lcx = cls(cost_fpath, features_fpath, resolution=resolution,
                  xmission_config=xmission_config,
                  min_line_length=min_line_length,
                  tb_layer_name=tb_layer_name,
                  iso_regions_layer_name=iso_regions_layer_name)

        licl = length_invariant_cost_layers
        least_costs = lcx.process_sc_points(capacity_class,
                                            cost_layers,
                                            sc_point_gids=sc_point_gids,
                                            nn_sinks=nn_sinks,
                                            clipping_buffer=clipping_buffer,
                                            barrier_mult=barrier_mult,
                                            max_workers=max_workers,
                                            save_paths=save_paths,
                                            radius=radius,
                                            expand_radius=expand_radius,
                                            mp_delay=mp_delay,
                                            simplify_geo=simplify_geo,
                                            length_invariant_cost_layers=licl,
                                            tracked_layers=tracked_layers,
                                            length_mult_kind=length_mult_kind,
                                            cell_size=cell_size)

        logger.info('{} connections were made to {} SC points in {:.4f} '
                    'minutes'
                    .format(len(least_costs),
                            len(least_costs['sc_point_gid'].unique()),
                            (time.time() - ts) / 60))

        return least_costs


class RegionalXmission(LeastCostXmission):
    """
    Compute Least Cost tie-line paths and full transmission cap cost
    for all supply curve points to all possible connections within the
    region containing the supply curve point.

    This class also offers a bit more flexibility than it's base
    (``LeastCostXmission``):

        - costs can be specified for region values of 0.
        - the ``trans_gid``, ``min_volt`` and ``max_volt`` columns are
          no longer required in the features input.

    """
    _TCC_CLASS = RegionalTransCapCosts

    def __init__(self, cost_fpath, features_fpath, regions,
                 region_identifier_column, resolution=RESOLUTION,
                 xmission_config=None, min_line_length=MINIMUM_SPUR_DIST_KM,
                 tb_layer_name=BARRIER_H5_LAYER_NAME,
                 iso_regions_layer_name=ISO_H5_LAYER_NAME):
        """
        Parameters
        ----------
        cost_fpath : str
            Path to h5 file with cost rasters and other required layers.
        features_fpath : str
            Path to GeoPackage with transmission features. All features
            except substations will be dropped. This table must have a
            `region_identifier` column which matches one of the
            `region_identifier` ID's in the `regions` GeoPackage.
        regions : str | GeoDataFrame
            Path to GeoPackage with regions or GeoDataFrame defining
            regions. This table must have the specified
            `region_identifier` column which matches the
            `region_identifier` column ID's in the `features_fpath`
            GeoPackage.
        region_identifier_column : str
            Name of column in `regions` GeoPackage containing a unique
            identifier for each region.
        resolution : int, optional
            SC point resolution. By default, ``128``.
        xmission_config : str | dict | XmissionConfig, optional
            Path to Xmission config .json, dictionary of Xmission config
            .jsons, or preloaded XmissionConfig objects.
            By default, ``None``.
        min_line_length : int | float, optional
            Minimum line length in km. By default, ``0``.
        tb_layer_name : str, default=:obj:`BARRIER_H5_LAYER_NAME`
            Name of transmission barrier layer in `cost_fpath` file.
            This layer defines the multipliers applied to the cost layer
            to determine LCP routing (but does not actually affect
            output costs). By default, :obj:`BARRIER_H5_LAYER_NAME`.
        iso_regions_layer_name : str, default=:obj:`ISO_H5_LAYER_NAME`
            Name of ISO regions layer in `cost_fpath` file. The layer
            maps pixels to ISO region ID's (1, 2, 3, 4, etc.) .
            By default, :obj:`ISO_H5_LAYER_NAME`.
            least cost tie-line path, by default 100
        """
        super().__init__(cost_fpath=cost_fpath,
                         features_fpath=features_fpath,
                         resolution=resolution,
                         xmission_config=xmission_config,
                         min_line_length=min_line_length,
                         tb_layer_name=tb_layer_name,
                         iso_regions_layer_name=iso_regions_layer_name)

        with ExclusionLayers(cost_fpath) as f:
            crs = CRS.from_string(f.crs)
        self._features = self._features.to_crs(crs)

        if isinstance(regions, str):
            regions = gpd.read_file(regions)
        self._regions = regions.to_crs(crs)
        self._rid_column = region_identifier_column
        self._region_features = {}

    @staticmethod
    def _load_trans_feats(features_fpath):
        """Load existing transmission features from disk. """

        logger.debug('Loading transmission features...')

        features = gpd.read_file(features_fpath)
        features = features.reset_index(drop=True)
        features = features.drop(columns=['bgid', 'egid', 'cap_left'],
                                 errors='ignore')
        mapping = {'gid': 'trans_gid'}
        features = features.rename(columns=mapping)
        mask = features["category"].isna()
        if mask.any():
            msg = f"Dropping {mask.sum():,} features with NaN category!"
            logger.warning(msg)
            warn(msg)
            features = features[~mask].reset_index(drop=True)

        return features, None

    def _clipped_features(self, rid):
        """Clip features to region and record the region ID. """
        if rid in self._region_features:
            return self._region_features[rid]

        mask = self._regions[self._rid_column] == rid
        sc_features = gpd.clip(self.features, self._regions[mask])
        self._region_features[rid] = sc_features
        return sc_features

    def _clip_to_sc_point(self, sc_point, *__,
                          clipping_buffer=CLIP_RASTER_BUFFER, radius=None,
                          expand_radius=True, **___):
        """Clip features to be substations in the region of the sc point.  """
        logger.debug('Clipping features to sc_point %d', sc_point.sc_point_gid)

        point = self.sc_points.loc[sc_point.name:sc_point.name].centroid
        map_func = region_mapper(self._regions, self._rid_column)
        rid = point.apply(map_func).values[0]
        logger.debug("  - Clipping features to region: %s", rid)
        sc_features = self._clipped_features(rid)
        logger.debug('%d transmission features found in clipped area ',
                     len(sc_features))

        with ExclusionLayers(self._cost_fpath) as f:
            transform = f.profile["transform"]

        resolution = transform[0]
        transform = rasterio.Affine(*transform)

        if radius is not None:
            sc_features = self._clip_to_radius(sc_point, radius, sc_features,
                                               clipping_buffer, resolution,
                                               expand_radius)

        if sc_features.empty:
            return sc_features, None

        radius = _compute_radius(sc_features, sc_point, transform,
                                 clipping_buffer)
        logger.debug('%d transmission features found in clipped area of '
                     'radius %.2f',
                     len(sc_features), radius)

        sc_features[self._rid_column] = rid
        return sc_features, radius

    @classmethod
    def run(cls, cost_fpath, features_fpath, regions_fpath,
            region_identifier_column, capacity_class, cost_layers,
            resolution=RESOLUTION, xmission_config=None,
            min_line_length=MINIMUM_SPUR_DIST_KM, sc_point_gids=None,
            clipping_buffer=CLIP_RASTER_BUFFER,
            tb_layer_name=BARRIER_H5_LAYER_NAME, barrier_mult=BARRIERS_MULT,
            iso_regions_layer_name=ISO_H5_LAYER_NAME, max_workers=None,
            simplify_geo=None, save_paths=False, radius=None,
            expand_radius=True, mp_delay=3, length_invariant_cost_layers=None,
            tracked_layers=None, length_mult_kind="linear",
            cell_size=CELL_SIZE):
        """
        Find Least Cost Transmission connections between desired
        sc_points and substations in their region.

        Parameters
        ----------
        cost_fpath : str
            Path to h5 file with cost rasters and other required layers
        features_fpath : str
            Path to GeoPackage with transmission features. All features
            except substations will be dropped. This table must have a
            `region_identifier` column which matches one of the
            `region_identifier` ID's in the `regions` GeoPackage.
        regions : str | GeoDataFrame
            Path to GeoPackage with regions or GeoDataFrame defining
            regions. This table must have the specified
            `region_identifier` column which matches the
            `region_identifier` column ID's in the `features_fpath`
            GeoPackage.
        region_identifier_column : str
            Name of column in `regions` GeoPackage containing a unique
            identifier for each region.
        capacity_class : str | int
            Capacity class of transmission features to connect supply
            curve points to.
        cost_layers : List[str]
            List of layers in H5 that are summed to determine total
            cost raster used for routing. Layer names may have curly
            brackets (``{}``), which will be filled in based on the
            `capacity_class` input (e.g. "tie_line_costs_{}MW").
        resolution : int, optional
            SC point resolution. By default, ``128``.
        xmission_config : str | dict | XmissionConfig, optional
            Path to Xmission config .json, dictionary of Xmission config
            .jsons, or preloaded XmissionConfig objects.
            By default, ``None``.
        min_line_length : int | float, optional
            Minimum line length in km. By default, ``0``.
        sc_point_gids : list, optional
            List of sc_point_gids to connect to. By default, ``None``,
            which processes all points.
        clipping_buffer : float, optional
            Buffer to expand clipping radius by. By default, ``1.05``.
        tb_layer_name : str, default=:obj:`BARRIER_H5_LAYER_NAME`
            Name of transmission barrier layer in `cost_fpath` file.
            This layer defines the multipliers applied to the cost layer
            to determine LCP routing (but does not actually affect
            output costs). By default, :obj:`BARRIER_H5_LAYER_NAME`.
        barrier_mult : int, optional
            Multiplier on transmission barrier costs.
            By default, ``100``.
        iso_regions_layer_name : str, default=:obj:`ISO_H5_LAYER_NAME`
            Name of ISO regions layer in `cost_fpath` file. The layer
            maps pixels to ISO region ID's (1, 2, 3, 4, etc.) .
            By default, :obj:`ISO_H5_LAYER_NAME`.
            least cost tie-line path, by default 100
        max_workers : int, optional
            Number of workers to use for processing. If 1 run in serial,
            if ``None`` use all available cores. By default, ``None``.
        simplify_geo : float | None, optional
            If float, simplify geometries using this value.
        save_paths : bool, optional
            Flag to save spur line path as a multi-line geometry.
            By default, ``False``.
        radius : None | int, optional
            Force clipping radius. Substations beyond this radius will
            not be considered for connection with supply curve point. If
            ``None``, no radius is forced, and connections to all
            available transmission features are computed.
            By default, ``None``.
        expand_radius : bool, optional
            Option to expand radius to include at least one connection
            feature. Has no effect if ``radius=None``.
            By default, ``True``.
        mp_delay : float, optional
            Delay in seconds between starting multi-process workers.
            Useful for reducing memory spike at working startup.
        length_invariant_cost_layers : List[str] | None, optional
            List of layers in H5 to be added to the 'base' greenfield
            cost raster. The costs specified by these layers are not
            scaled with distance traversed across the cell (i.e. fixed
            one-time costs for crossing these cells).
        tracked_layers : dict, optional
            Dictionary mapping layer names to strings, where the strings
            are numpy methods that should be applied to the layer along
            the LCP. For example,
            ``tracked_layers={'layer_1': 'mean', 'layer_2': 'max}``
            would report the average of ``layer_1`` values along the
            least cost path and the max of ``layer_2`` values along the
            least cost path. Examples of numpy methods (non-exhaustive):

                - mean
                - max
                - min
                - mode
                - median
                - std

            By default, ``None``, which does not track any extra layers.
        length_mult_kind : {"step", "linear"}, default="linear"
            Type of length multiplier calcualtion. "step" computes
            length multipliers using a step function, while "linear"
            computes the length multiplier using a linear interpolation
            between 0 amd 10 mile spur-line lengths.
            By default, ``"linear"``.
        cell_size : int, optional
            Side length of each cell, in meters. Cells are assumed to be
            square. By default, :obj:`CELL_SIZE`.

        Returns
        -------
        least_costs : pandas.DataFrame | gpd.DataFrame
            Least cost connections between all supply curve points and
            the substations in their region with the given capacity
            class.
        """
        ts = time.time()
        lcx = cls(cost_fpath, features_fpath, regions_fpath,
                  region_identifier_column, resolution, xmission_config,
                  min_line_length, tb_layer_name=tb_layer_name,
                  iso_regions_layer_name=iso_regions_layer_name)
        licl = length_invariant_cost_layers
        least_costs = lcx.process_sc_points(capacity_class, cost_layers,
                                            sc_point_gids=sc_point_gids,
                                            clipping_buffer=clipping_buffer,
                                            barrier_mult=barrier_mult,
                                            max_workers=max_workers,
                                            save_paths=save_paths,
                                            radius=radius,
                                            expand_radius=expand_radius,
                                            mp_delay=mp_delay,
                                            simplify_geo=simplify_geo,
                                            length_invariant_cost_layers=licl,
                                            tracked_layers=tracked_layers,
                                            length_mult_kind=length_mult_kind,
                                            cell_size=cell_size)

        logger.info('{} connections were made to {} SC points in {:.4f} '
                    'minutes'
                    .format(len(least_costs),
                            len(least_costs['sc_point_gid'].unique()),
                            (time.time() - ts) / 60))

        return least_costs


def region_mapper(regions, region_identifier_column):
    """Generate a function to map points to a region.

    The returned mapping function maps a point to a unique value from
    the `region_identifier_column` column in the input GeoPackage.

    Parameters
    ----------
    regions : gpd.GeoPackage
        GeoPackage defining the regions. This table must
        have a `region_identifier_column` column which uniquely
        identifies the region, as well as a geometry for each region.
    """
    def _map_region(point):
        """Find the region ID for the input point. """
        idx = regions.distance(point).sort_values().index[0]
        return regions.loc[idx, region_identifier_column]

    return _map_region


def _collect_future_chunks(futures, least_cost_paths):
    """Collect all futures from the input dictionary. """

    num_to_collect = len(futures)
    for i, future in enumerate(as_completed(futures), start=1):
        futures.pop(future)
        sc_costs = future.result()
        if sc_costs is not None:
            least_cost_paths.append(sc_costs)
        logger.debug('Collected {} of {} futures!'.format(i, num_to_collect))
        log_mem(logger)

    return least_cost_paths


def _starting_costs(cost_fpath, points, cost_layers,
                    length_invariant_cost_layers, barrier_mult,
                    tb_layer_name=BARRIER_H5_LAYER_NAME):
    """Extract the starting point cost"""
    costs = 0
    rows, cols = points["row"].values, points["col"].values
    with ExclusionLayers(cost_fpath) as f:
        for layer in (cost_layers or []):
            costs += f[layer][rows, cols]

        for layer in (length_invariant_cost_layers or []):
            costs += f[layer][rows, cols]

        barrier = f[tb_layer_name, rows, cols]
        barrier = barrier * float(barrier_mult)
        costs *= 1 + barrier
        costs = np.where(costs <= 0, -1, costs)

    return costs


def _compute_radius(sc_features, sc_point, transform, clipping_buffer):
    """Compute boundary clipping radius based on feature total bounds. """
    minx, miny, maxx, maxy = sc_features.total_bounds
    r1, c1 = rasterio.transform.rowcol(transform, minx, miny)
    r2, c2 = rasterio.transform.rowcol(transform, maxx, maxy)
    bounds = pd.DataFrame({"row": [r1, r2], "col": [c1, c2]})
    dists = bounds[['row', 'col']] - sc_point[['row', 'col']]
    return int(np.ceil(dists.abs().values.max() * clipping_buffer))



# -*- coding: utf-8 -*-
"""
Module to compute least cost transmission paths and distances
"""
from concurrent.futures import as_completed
import geopandas as gpd
import logging
import numpy as np
import os
import pandas as pd
from pyproj.crs import CRS
import rasterio
import time
from warnings import warn

from reV.handlers.exclusions import ExclusionLayers
from rex.utilities.execution import SpawnProcessPool
from rex.utilities.loggers import log_mem

from reVX.handlers.layered_h5 import crs_match
from reVX.least_cost_xmission.config.constants import (CELL_SIZE,
                                                       TRANS_LINE_CAT,
                                                       BARRIERS_MULT,
                                                       BARRIER_H5_LAYER_NAME)
from reVX.least_cost_xmission.trans_cap_costs import (TieLineCosts,
                                                      ReinforcementLineCosts)

logger = logging.getLogger(__name__)


class LeastCostPaths:
    """
    Compute least cost paths between desired locations
    """

    def __init__(self, cost_fpath, features_fpath, clip_buffer=0,
                 tb_layer_name=BARRIER_H5_LAYER_NAME):
        """
        Parameters
        ----------
        cost_fpath : str
            Path to h5 file with cost rasters and other required layers
        features_fpath : str
            Path to GeoPackage with transmission features
        resolution : int, optional
            SC point resolution, by default 128
        clip_buffer : int, optional
            Optional number of array elements to buffer clip area by.
            By default, ``0``.
        tb_layer_name : str, default=:obj:`BARRIER_H5_LAYER_NAME`
            Name of transmission barrier layer in `cost_fpath` file.
            This layer defines the multipliers applied to the cost layer
            to determine LCP routing (but does not actually affect
            output costs). By default, :obj:`BARRIER_H5_LAYER_NAME`.
        """
        self._cost_fpath = cost_fpath
        self._tb_layer_name = tb_layer_name
        self._check_layers()

        out = self._map_to_costs(cost_fpath, gpd.read_file(features_fpath),
                                 clip_buffer=clip_buffer)
        self._features, self._row_slice, self._col_slice, self._shape = out
        self._features = self._features.drop(columns='geometry')
        self._start_feature_ind = 0

        logger.debug('{} initialized'.format(self))

    def __repr__(self):
        msg = ("{} to be computed for {:,d} features"
               .format(self.__class__.__name__, len(self._features)))
        return msg

    @property
    def features(self):
        """
        Table of features to compute paths for

        Returns
        -------
        pandas.DataFrame
        """
        return self._features

    def _check_layers(self):
        """Check to make sure the required layers are in cost_fpath. """
        self._check_layers_in_h5(self._cost_fpath, [self._tb_layer_name])

    @classmethod
    def _check_layers_in_h5(cls, cost_fpath, layers):
        """
        Check to make sure the required layers are in cost_fpath

        Parameters
        ----------
        cost_fpath : str
            Path to h5 file with cost rasters and other required layers
        layers : iter
            Iterable of required layer names.

        Raises
        ------
        RuntimeError
            If the required layers are not found in the cost H5.
        """
        with ExclusionLayers(cost_fpath) as f:
            missing = []
            for lyr in layers:
                if lyr not in f:
                    missing.append(lyr)

            if missing:
                msg = ("The following layers are required to compute Least "
                       "Cost Transmission but are missing from {}:\n{}"
                       .format(cost_fpath, missing))
                logger.error(msg)
                raise RuntimeError(msg)

    @staticmethod
    def _get_feature_cost_indices(features, crs, transform, shape):
        """
        Map features to cost row, col indices using rasterio transform

        Parameters
        ----------
        features : gpd.GeoDataFrame
            GeoDataFrame of features to map to cost raster
        crs : pyproj.crs.CRS
            CRS of cost raster
        transform : raster.Affine
            Transform of cost raster
        shape : tuple
            Cost raster shape

        Returns
        -------
        row : ndarray
            Vector of row indices for each feature
        col : ndarray
            Vector of col indices for each features
        mask : ndarray
            Boolean mask of features with indices outside of cost raster
        """
        feat_crs = features.crs.to_dict()
        cost_crs = crs.to_dict()
        if not crs_match(cost_crs, feat_crs):
            msg = ('input crs ({}) does not match cost raster crs ({}) '
                   'and will be transformed!'.format(feat_crs, cost_crs))
            logger.warning(msg)
            warn(msg)
            features = features.to_crs(crs)

        logger.debug('Map %d features to cost raster', len(features))
        logger.debug('First few features:\n%s', str(features.head()))
        logger.debug('Transform:\n%s', str(transform))
        coords = features['geometry'].centroid
        row, col = rasterio.transform.rowcol(transform, coords.x.values,
                                             coords.y.values)
        logger.debug('Mapping done!')
        row = np.array(row)
        col = np.array(col)

        # Remove features outside of the cost domain
        mask = row >= 0
        mask &= row < shape[0]
        mask &= col >= 0
        mask &= col < shape[1]

        logger.debug('Mask computed!')
        return row, col, mask

    @staticmethod
    def _get_clip_slice(row, col, shape, clip_buffer=0):
        """
        Clip cost raster to bounds of features

        Parameters
        ----------
        row : ndarray
            Vector of row indices
        col : ndarray
            Vector of col indices
        shape : tuple
            Full cost array shape
        clip_buffer : int, optional
            Optional number of array elements to buffer clip area by.
            By default, ``0``.

        Returns
        -------
        row_slice : slice
            Row slice to clip too
        col_slice : slice
            Col slice to clip too
        """
        row_slice = slice(max(row.min() - 1 - clip_buffer, 0),
                          min(row.max() + 1 + clip_buffer, shape[0]))
        col_slice = slice(max(col.min() - 1 - clip_buffer, 0),
                          min(col.max() + 1 + clip_buffer, shape[1]))

        return row_slice, col_slice

    @classmethod
    def _map_to_costs(cls, cost_fpath, features, clip_buffer=0):
        """
        Map features to cost arrays

        Parameters
        ----------
        cost_fpath : str
            Path to h5 file with cost rasters and other required layers
        features : gpd.GeoDataFrame
            GeoDataFrame of features to connect

        Returns
        -------
        features : gpd.GeoDataFrame
            Table of features to compute LeastCostPaths for
        row_slice : slice
            Clipping slice along axis-0 (rows)
        col_slice : slice
            Clipping slice along axis-1 (cols)
        shape : tuple
            Full cost raster shape
        clip_buffer : int, optional
            Optional number of array elements to buffer clip area by.
            By default, ``0``.
        """
        with ExclusionLayers(cost_fpath) as f:
            crs = CRS.from_string(f.crs)
            transform = rasterio.Affine(*f.profile['transform'])
            shape = f.shape

        row, col, mask = cls._get_feature_cost_indices(features, crs,
                                                       transform, shape)
        if any(~mask):
            msg = ("The following features are outside of the cost exclusion "
                   "domain and will be dropped:\n{}"
                   .format(features.loc[~mask]))
            logger.warning(msg)
            warn(msg)
            row = row[mask]
            col = col[mask]
            features = features.loc[mask].reset_index(drop=True)

        logger.debug('Getting clip size...')
        row_slice, col_slice = cls._get_clip_slice(row, col, shape,
                                                   clip_buffer=clip_buffer)
        logger.debug('Done!')

        features['row'] = row - row_slice.start
        features['col'] = col - col_slice.start

        return features, row_slice, col_slice, shape

    @property
    def start_indices(self):
        """
        Tuple of (row_idx, col_idx) in the cost array indicating the
        start position of all paths to compute (typically, this is
        the centroid of the supply curve cell under consideration).
        Paths will be computed from this start location to each of the
        `end_indices`, which are also locations in the cost array
        (typically transmission feature locations).

        Returns
        -------
        tuple
        """
        start_indices = (self._features.loc[self._start_feature_ind,
                                            ['row', 'col']].values)
        if len(start_indices.shape) == 2:
            start_indices = start_indices[0]
        return start_indices

    @property
    def end_features(self):
        """
        GeoDataFrame containing the transmission features to compute the
        least cost paths to, starting from the `start_indices`
        (typically the centroid of the supply curve cell under
        consideration).

        Returns
        -------
        pandas.DataFrame
        """
        end_features = self._features.drop(index=self._start_feature_ind)
        end_features['start_index'] = self._start_feature_ind
        return end_features.reset_index(drop=False)

    @property
    def end_indices(self):
        """
        Tuple (row, col) index or list of (row, col) indices in the
        cost array indicating the end location(s) to compute least
        cost paths to (typically transmission feature locations).
        Paths are computed from the `start_indices` (typically the
        centroid of the supply curve cell under consideration) to each
        of the individual pairs of `end_indices`.

        Returns
        -------
        tuple | list
        """
        return self.end_features[['row', 'col']].values

    def process_least_cost_paths(self, cost_layers, barrier_mult=BARRIERS_MULT,
                                 indices=None, max_workers=None,
                                 save_paths=False,
                                 length_invariant_cost_layers=None,
                                 tracked_layers=None, cell_size=CELL_SIZE):
        """
        Find Least Cost Paths between all pairs of provided features for
        the given tie-line capacity class

        Parameters
        ----------
        cost_layers : List[str]
            List of layers in H5 that are summed to determine total
            costs raster used for routing. Costs and distances for each
            individual layer are also reported (e.g. wet and dry costs).
            deteremining path using main cost layer.
        barrier_mult : int, optional
            Transmission barrier multiplier, used when computing the
            least cost tie-line path, by default 100
        indices : iterable, optonal
            Indices of the transmission features input that should be
            processed. By default ``None``, which process all
            transmission features.
        max_workers : int, optional
            Number of workers to use for processing, if 1 run in serial,
            if None use all available cores, by default None
        save_paths : bool, optional
            Flag to save least cost path as a multi-line geometry,
            by default False
        length_invariant_cost_layers : List[str] | None, optional
            List of layers in H5 to be added to the cost raster. The
            costs specified by these layers are not scaled with distance
            traversed across the cell (i.e. fixed one-time costs for
            crossing these cells).
        tracked_layers : dict, optional
            Dictionary mapping layer names to strings, where the strings
            are numpy methods that should be applied to the layer along
            the LCP. For example,
            ``tracked_layers={'layer_1': 'mean', 'layer_2': 'max}``
            would report the average of ``layer_1`` values along the
            least cost path and the max of ``layer_2`` values along the
            least cost path. Examples of numpy methods (non-exhaustive):

                - mean
                - max
                - min
                - mode
                - median
                - std

            By default, ``None``, which does not track any extra layers.
        cell_size : int, optional
            Side length of each cell, in meters. Cells are assumed to be
            square. By default, :obj:`CELL_SIZE`.

        Returns
        -------
        least_cost_paths : pandas.DataFrame | gpd.GeoDataFrame
            DataFrame of lengths and costs for each path or GeoDataFrame
            of length, cost, and geometry for each path
        """
        max_workers = os.cpu_count() if max_workers is None else max_workers
        indices = self.features.index if indices is None else indices
        least_cost_paths = []
        if max_workers > 1:
            logger.info('Computing Least Cost Paths in parallel on {} workers'
                        .format(max_workers))
            log_mem(logger)
            loggers = [__name__, 'reV', 'reVX']
            with SpawnProcessPool(max_workers=max_workers,
                                  loggers=loggers) as exe:
                least_cost_paths = self._compute_paths_in_chunks(
                    exe, max_workers, indices, cost_layers, barrier_mult,
                    save_paths, length_invariant_cost_layers, tracked_layers,
                    cell_size=cell_size)
        else:
            least_cost_paths = []
            logger.info('Computing Least Cost Paths in serial')
            log_mem(logger)
            licl = length_invariant_cost_layers
            for ind, start in enumerate(indices, start=1):
                self._start_feature_ind = start
                lcp = TieLineCosts.run(self._cost_fpath,
                                       self.start_indices, self.end_indices,
                                       cost_layers,
                                       self._row_slice, self._col_slice,
                                       tb_layer_name=self._tb_layer_name,
                                       barrier_mult=barrier_mult,
                                       save_paths=save_paths,
                                       length_invariant_cost_layers=licl,
                                       tracked_layers=tracked_layers,
                                       cell_size=cell_size)
                end_features = self.end_features.drop(columns=['row', 'col'],
                                                      errors="ignore")
                lcp = pd.concat((lcp, end_features), axis=1)
                least_cost_paths.append(lcp)

                logger.debug('Least cost path {} of {} complete!'
                             .format(ind, len(self.features)))
                log_mem(logger)

        least_cost_paths = pd.concat(least_cost_paths, ignore_index=True)

        return least_cost_paths

    def _compute_paths_in_chunks(self, exe, max_submissions, indices,
                                 cost_layers, barrier_mult, save_paths,
                                 licl, tracked_layers, cell_size):
        """Compute LCP's in parallel using futures. """
        futures, paths = {}, []

        for ind, start in enumerate(indices, start=1):
            self._start_feature_ind = start
            future = exe.submit(TieLineCosts.run, self._cost_fpath,
                                self.start_indices, self.end_indices,
                                cost_layers,
                                self._row_slice, self._col_slice,
                                tb_layer_name=self._tb_layer_name,
                                barrier_mult=barrier_mult,
                                save_paths=save_paths,
                                length_invariant_cost_layers=licl,
                                tracked_layers=tracked_layers,
                                cell_size=cell_size)
            futures[future] = self.end_features
            logger.debug('Submitted {} of {} futures'
                         .format(ind, len(indices)))
            log_mem(logger)
            if ind % max_submissions == 0:
                paths = _collect_future_chunks(futures, paths)
        paths = _collect_future_chunks(futures, paths)
        return paths

    @classmethod
    def run(cls, cost_fpath, features_fpath, cost_layers,
            clip_buffer=0, tb_layer_name=BARRIER_H5_LAYER_NAME,
            barrier_mult=BARRIERS_MULT, indices=None, max_workers=None,
            save_paths=False, length_invariant_cost_layers=None,
            tracked_layers=None, cell_size=CELL_SIZE):
        """
        Find Least Cost Paths between all pairs of provided features for
        the given tie-line capacity class

        Parameters
        ----------
        cost_fpath : str
            Path to h5 file with cost rasters and other required layers
        features_fpath : str
            Path to GeoPackage with transmission features
        cost_layers : List[str]
            List of layers in H5 that are summed to determine total
            costs raster used for routing. Costs and distances for each
            individual layer are also reported (e.g. wet and dry costs).
            deteremining path using main cost layer.
        clip_buffer : int, optional
            Optional number of array elements to buffer clip area by.
            By default, ``0``.
        tb_layer_name : str, default=:obj:`BARRIER_H5_LAYER_NAME`
            Name of transmission barrier layer in `cost_fpath` file.
            This layer defines the multipliers applied to the cost layer
            to determine LCP routing (but does not actually affect
            output costs). By default, :obj:`BARRIER_H5_LAYER_NAME`.
        barrier_mult : int, optional
            Transmission barrier multiplier, used when computing the
            least cost tie-line path, by default 100
        indices : iterable, optonal
            Indices of the transmission features input that should be
            processed. By default ``None``, which process all
            transmission features.
        max_workers : int, optional
            Number of workers to use for processing, if 1 run in serial,
            if None use all available cores, by default None
        save_paths : bool, optional
            Flag to save least cost path as a multi-line geometry,
            by default False
        length_invariant_cost_layers : List[str] | None, optional
            List of layers in H5 to be added to the cost raster. The
            costs specified by these layers are not scaled with distance
            traversed across the cell (i.e. fixed one-time costs for
            crossing these cells).
        tracked_layers : dict, optional
            Dictionary mapping layer names to strings, where the strings
            are numpy methods that should be applied to the layer along
            the LCP. For example,
            ``tracked_layers={'layer_1': 'mean', 'layer_2': 'max}``
            would report the average of ``layer_1`` values along the
            least cost path and the max of ``layer_2`` values along the
            least cost path. Examples of numpy methods (non-exhaustive):

                - mean
                - max
                - min
                - mode
                - median
                - std

            By default, ``None``, which does not track any extra layers.
        cell_size : int, optional
            Side length of each cell, in meters. Cells are assumed to be
            square. By default, :obj:`CELL_SIZE`.

        Returns
        -------
        least_cost_paths : pandas.DataFrame | gpd.GeoDataFrame
            DataFrame of lengths and costs for each path or GeoDataFrame
            of length, cost, and geometry for each path
        """
        ts = time.time()
        lcp = cls(cost_fpath, features_fpath, clip_buffer=clip_buffer,
                  tb_layer_name=tb_layer_name)
        least_cost_paths = lcp.process_least_cost_paths(
            cost_layers,
            barrier_mult=barrier_mult,
            indices=indices,
            save_paths=save_paths,
            max_workers=max_workers,
            length_invariant_cost_layers=length_invariant_cost_layers,
            tracked_layers=tracked_layers, cell_size=cell_size)

        logger.info('{} paths were computed in {:.4f} hours'
                    .format(len(least_cost_paths),
                            (time.time() - ts) / 3600))

        return least_cost_paths


class ReinforcementPaths(LeastCostPaths):
    """
    Compute reinforcement line paths between substations and a single
    balancing area network node.
    """
    def __init__(self, cost_fpath, features, transmission_lines,
                 clip_buffer=0, tb_layer_name=BARRIER_H5_LAYER_NAME):
        """
        Parameters
        ----------
        cost_fpath : str
            Path to h5 file with cost rasters and other required layers.
        features : geopandas.GeoPackage
            GeoPackage with transmission features. The network
            node must be the first row of the GeoPackage - the rest
            should be substations that need to connect to that node.
        transmission_lines :dict
            Dictionary where the keys are the names of cost layers in
            the cost HDF5 file and values are arrays with the
            corresponding existing transmission lines rastered into
            them (i.e. array value is 1 at a pixel if there is a
            transmission line, otherwise 0). These arrays will be used
            to compute the reinforcement costs along existing
            transmission lines of differing voltages.
        clip_buffer : int, optional
            Optional number of array elements to buffer clip area by.
            By default, ``0``.
        tb_layer_name : str, default=:obj:`BARRIER_H5_LAYER_NAME`
            Name of transmission barrier layer in `cost_fpath` file.
            This layer defines the multipliers applied to the cost layer
            to determine LCP routing (but does not actually affect
            output costs). By default, :obj:`BARRIER_H5_LAYER_NAME`.
        """
        self._cost_fpath = cost_fpath
        self._tb_layer_name = tb_layer_name
        self._check_layers()

        self._features, self._row_slice, self._col_slice, self._shape = \
            self._map_to_costs(cost_fpath, features, clip_buffer)
        self._features = self._features.drop(columns='geometry')

        network_node = (self._features.iloc[0:1]
                        .dropna(axis="columns", how="all"))
        self._start_indices = network_node[['row', 'col']].values[0]
        self._features = (self._features.iloc[1:]
                          .reset_index(drop=True)
                          .dropna(axis="columns", how="all"))

        self._transmission_lines = {
            capacity_mw: lines[self._row_slice, self._col_slice]
            for capacity_mw, lines in transmission_lines.items()
        }

    @property
    def start_indices(self):
        """
        Tuple of (row_idx, col_idx) in the cost array indicating the
        start position of all reinforcement line paths to compute
        (typically, this is the location of the network node in the
        reinforcement region). Paths will be computed from this start
        location to each of the `end_indices`, which are also locations
        in the cost array (typically substations within the
        reinforcement region of the network node).

        Returns
        -------
        tuple
        """
        return self._start_indices

    @property
    def end_indices(self):
        """
        Tuple (row, col) index or list of (row, col) indices in the cost
        array indicating the end location(s) to compute reinforcement
        line paths to (typically substations within a single
        reinforcement region). Paths are computed from the
        `start_indices` (typically the network node of the reinforcement
        region) to each of the individual pairs of `end_indices`.

        Returns
        -------
        tuple | list
        """
        return self._features[['row', 'col']].values

    def process_least_cost_paths(self, capacity_class, cost_layers,
                                 barrier_mult=BARRIERS_MULT, max_workers=None,
                                 save_paths=False,
                                 length_invariant_cost_layers=None,
                                 tracked_layers=None, cell_size=CELL_SIZE):
        """
        Find the reinforcement line paths between the network node and
        the substations for the given tie-line capacity class

        Parameters
        ----------
        capacity_class : int | str
            Capacity class of the 'base' greenfield costs layer. Costs
            will be scaled by the capacity corresponding to this class
            to report reinforcement costs as $/MW.
        cost_layers : List[str]
            List of layers in H5 that are summed to determine total
            'base' greenfield costs raster used for routing. 'Base'
            greenfield costs are only used if the reinforcement path
            *must* deviate from existing transmission lines. Typically,
            a capacity class of 400 MW (230kV transmission line) is used
            for the base greenfield costs.
        barrier_mult : int, optional
            Multiplier on transmission barrier costs.
            By default, ``100``.
        max_workers : int, optional
            Number of workers to use for processing. If 1 run in serial,
            if ``None`` use all available cores. By default, ``None``.
        save_paths : bool, optional
            Flag to save reinforcement line path as a multi-line
            geometry. By default, ``False``.
        length_invariant_cost_layers : List[str] | None, optional
            List of layers in H5 to be added to the 'base' greenfield
            cost raster. The costs specified by these layers are not
            scaled with distance traversed across the cell (i.e. fixed
            one-time costs for crossing these cells).
        tracked_layers : dict, optional
            Dictionary mapping layer names to strings, where the strings
            are numpy methods that should be applied to the layer along
            the LCP. For example,
            ``tracked_layers={'layer_1': 'mean', 'layer_2': 'max}``
            would report the average of ``layer_1`` values along the
            least cost path and the max of ``layer_2`` values along the
            least cost path. Examples of numpy methods (non-exhaustive):

                - mean
                - max
                - min
                - mode
                - median
                - std

            By default, ``None``, which does not track any extra layers.
        cell_size : int, optional
            Side length of each cell, in meters. Cells are assumed to be
            square. By default, :obj:`CELL_SIZE`.

        Returns
        -------
        least_cost_paths : pandas.DataFrame | gpd.GeoDataFrame
            DataFrame of lengths and costs for each reinforcement line
            path or GeoDataFrame of length, cost, and geometry for each
            reinforcement line path.
        """

        logger.info('Computing reinforcement path costs for start index {}'
                    .format(self._start_indices))
        log_mem(logger)
        max_workers = os.cpu_count() if max_workers is None else max_workers
        max_workers = int(max_workers)

        licl = length_invariant_cost_layers
        least_cost_paths = []
        if max_workers > 1:
            logger.info('Computing Reinforcement Cost Paths in parallel on '
                        '%d workers', max_workers)
            log_mem(logger)
            loggers = [__name__, 'reV', 'reVX']
            with SpawnProcessPool(max_workers=max_workers,
                                  loggers=loggers) as exe:

                futures = {}
                for ind in range(max_workers):
                    feats = self._features.iloc[ind::max_workers]
                    future = exe.submit(ReinforcementLineCosts.run,
                                        self._transmission_lines,
                                        self._cost_fpath,
                                        self.start_indices,
                                        feats[['row', 'col']].values,
                                        capacity_class,
                                        cost_layers,
                                        self._row_slice,
                                        self._col_slice,
                                        tb_layer_name=self._tb_layer_name,
                                        barrier_mult=barrier_mult,
                                        save_paths=save_paths,
                                        length_invariant_cost_layers=licl,
                                        tracked_layers=tracked_layers,
                                        cell_size=cell_size)
                    futures[future] = feats
                    logger.debug('Submitted %d of %d futures',
                                 ind + 1, max_workers)
                    log_mem(logger)

                least_cost_paths = _collect_future_chunks(futures,
                                                          least_cost_paths)
                least_cost_paths = pd.concat(least_cost_paths,
                                             ignore_index=True)

        else:
            lcp = ReinforcementLineCosts.run(self._transmission_lines,
                                             self._cost_fpath,
                                             self.start_indices,
                                             self.end_indices,
                                             capacity_class,
                                             cost_layers,
                                             self._row_slice,
                                             self._col_slice,
                                             tb_layer_name=self._tb_layer_name,
                                             barrier_mult=barrier_mult,
                                             save_paths=save_paths,
                                             length_invariant_cost_layers=licl,
                                             tracked_layers=tracked_layers,
                                             cell_size=cell_size)
            least_cost_paths = lcp.merge(self._features, on=["row", "col"])

        least_cost_paths = least_cost_paths.dropna(axis="columns", how="all")
        drop_cols = ['index', 'row', 'col']
        return least_cost_paths.drop(columns=drop_cols, errors="ignore")

    @classmethod
    def run(cls, cost_fpath, features_fpath, network_nodes_fpath,
            region_identifier_column, transmission_lines_fpath,
            capacity_class, cost_layers, xmission_config=None, clip_buffer=0,
            tb_layer_name=BARRIER_H5_LAYER_NAME, barrier_mult=BARRIERS_MULT,
            indices=None, max_workers=None, save_paths=False,
            length_invariant_cost_layers=None, tracked_layers=None,
            ss_id_col="poi_gid", cell_size=CELL_SIZE):
        """
        Find the reinforcement line paths between the network node and
        the substations for the given tie-line capacity class

        Parameters
        ----------
        cost_fpath : str
            Path to h5 file with cost rasters and other required layers.
        features_fpath : str
            Path to GeoPackage with transmission features. The network
            node must be the first row of the GeoPackage - the rest
            should be substations that need to connect to that node.
            This table must have a `region_identifier_column` column
            which matches the `region_identifier_column` ID of the
            network node to the `region_identifier_column` ID of the
            substations that should connect to it.
        network_nodes_fpath : str
            Path to GeoPackage with network node endpoints. The
            endpoints should have a `region_identifier_column` column
            that identifies matches exactly one of the ID's in the
            reinforcement regions GeoPackage to be used in downstream
            models.
        region_identifier_column : str
            Name of column in `network_nodes_fpath` GeoPackage
            containing a unique identifier for each region.
        capacity_class : int | str
            Capacity class of the 'base' greenfield costs layer. Costs
            will be scaled by the capacity corresponding to this class
            to report reinforcement costs as $/MW.
        cost_layers : List[str]
            List of layers in H5 that are summed to determine total
            'base' greenfield costs raster used for routing. 'Base'
            greenfield costs are only used if the reinforcement path
            *must* deviate from existing transmission lines. Typically,
            a capacity class of 400 MW (230kV transmission line) is used
            for the base greenfield costs.
        xmission_config : str | dict | XmissionConfig, optional
            Path to Xmission config .json, dictionary of Xmission config
            .jsons, or preloaded XmissionConfig objects.
            By default, ``None``.
        clip_buffer : int, optional
            Optional number of array elements to buffer clip area by.
            By default, ``0``.
        tb_layer_name : str, default=:obj:`BARRIER_H5_LAYER_NAME`
            Name of transmission barrier layer in `cost_fpath` file.
            This layer defines the multipliers applied to the cost layer
            to determine LCP routing (but does not actually affect
            output costs). By default, :obj:`BARRIER_H5_LAYER_NAME`.
        barrier_mult : int, optional
            Multiplier on transmission barrier costs.
            By default, ``100``.
        indices : iterable, optonal
            Indices corresponding to the network nodes that should be
            processed. By default ``None``, which process all network
            nodes.
        max_workers : int, optional
            Number of workers to use for processing. If 1 run in serial,
            if ``None`` use all available cores. By default, ``None``.
        save_paths : bool, optional
            Flag to save reinforcement line path as a multi-line
            geometry. By default, ``False``.
        length_invariant_cost_layers : List[str] | None, optional
            List of layers in H5 to be added to the 'base' greenfield
            cost raster. The costs specified by these layers are not
            scaled with distance traversed across the cell (i.e. fixed
            one-time costs for crossing these cells).
        tracked_layers : dict, optional
            Dictionary mapping layer names to strings, where the strings
            are numpy methods that should be applied to the layer along
            the LCP. For example,
            ``tracked_layers={'layer_1': 'mean', 'layer_2': 'max}``
            would report the average of ``layer_1`` values along the
            least cost path and the max of ``layer_2`` values along the
            least cost path. Examples of numpy methods (non-exhaustive):

                - mean
                - max
                - min
                - mode
                - median
                - std

            By default, ``None``, which does not track any extra layers.
        ss_id_col : str, default="poi_gid"
            Name of column containing unique identifier for each
            substation. This column will be used to compute minimum
            reinforcement cost per substation.
            By default, ``"poi_gid"``.
        cell_size : int, optional
            Side length of each cell, in meters. Cells are assumed to be
            square. By default, :obj:`CELL_SIZE`.

        Returns
        -------
        least_cost_paths : pandas.DataFrame | gpd.GeoDataFrame
            DataFrame of lengths and costs for each reinforcement line
            path or GeoDataFrame of length, cost, and geometry for each
            reinforcement line path.
        """
        ts = time.time()
        least_cost_paths = []
        licl = length_invariant_cost_layers
        lcp_kwargs = {"capacity_class": capacity_class,
                      "cost_layers": cost_layers,
                      "length_invariant_cost_layers": licl,
                      "barrier_mult": barrier_mult,
                      "save_paths": save_paths,
                      "max_workers": max_workers,
                      "tracked_layers": tracked_layers,
                      "cell_size": cell_size}
        with ExclusionLayers(cost_fpath) as f:
            cost_crs = CRS.from_string(f.crs)
            cost_shape = f.shape
            cost_transform = rasterio.Affine(*f.profile['transform'])

        logger.info('Loading features from %s', features_fpath)
        features = gpd.read_file(features_fpath).to_crs(cost_crs)
        mapping = {'gid': ss_id_col}
        substations = features.rename(columns=mapping)
        substations = substations.dropna(axis="columns", how="all")
        logger.info('Loaded %d features from %s', len(substations),
                    features_fpath)

        logger.info('Loading tline shapes from %s', transmission_lines_fpath)
        lines = gpd.read_file(transmission_lines_fpath).to_crs(cost_crs)
        mapping = {'VOLTAGE': 'voltage'}
        lines = lines.rename(columns=mapping)
        transmission_lines = (lines[lines.category == TRANS_LINE_CAT]
                              .reset_index(drop=True))
        logger.info('Loaded %d tline shapes from %s', len(transmission_lines),
                    transmission_lines_fpath)

        logger.debug("Rasterizing transmission lines onto grid...")
        transmission_lines = _rasterize_transmission(transmission_lines,
                                                     xmission_config,
                                                     cost_shape,
                                                     cost_transform)

        logger.info('Loading network nodes from %s', network_nodes_fpath)
        network_nodes = gpd.read_file(network_nodes_fpath).to_crs(cost_crs)
        indices = network_nodes.index if indices is None else indices
        for loop_ind, index in enumerate(indices, start=1):
            network_node = (network_nodes.iloc[index:index + 1]
                            .reset_index(drop=True))
            rid = network_node[region_identifier_column].values[0]
            mask = substations[region_identifier_column] == rid
            logger.debug('Computing reinfocements to %s in region %s',
                         str(network_node), rid)
            node_substations = substations[mask].reset_index(drop=True)
            if len(node_substations) == 0:
                continue

            logger.debug('Found %d unique %s',
                         node_substations[ss_id_col].unique().shape[0],
                         ss_id_col)
            logger.info('Working on %d substations in region %s',
                        len(node_substations), rid)
            node_features = pd.concat([network_node, node_substations])
            rp = cls(cost_fpath, node_features, transmission_lines,
                     clip_buffer=clip_buffer, tb_layer_name=tb_layer_name)
            node_least_cost_paths = rp.process_least_cost_paths(**lcp_kwargs)
            node_least_cost_paths[region_identifier_column] = rid
            least_cost_paths += [node_least_cost_paths]

            logger.debug('Found %d reinforcement paths for network node',
                         len(node_least_cost_paths))
            logger.debug('Computed reinforcements for {}/{} network nodes'
                         .format(loop_ind, len(indices)))

        logger.info('Paths to {} network node(s) were computed in {:.4f} hours'
                    .format(len(least_cost_paths), (time.time() - ts) / 3600))

        costs = pd.concat(least_cost_paths, ignore_index=True)
        logger.debug('Computed %d reinforcement paths for %d unique POIs',
                     len(costs), costs[ss_id_col].unique().shape[0])
        return min_reinforcement_costs(costs, group_col=ss_id_col)


def _rasterize_transmission(transmission_lines, xmission_config, cost_shape,
                            cost_transform):
    """Rasterize transmission lines and assemble them into a dict. """

    transmission_lines_dict = {}
    capacities = sorted(xmission_config['power_classes'].values())
    v_min = 0
    for capacity in capacities[:-1]:
        v_max = xmission_config['power_to_voltage'][str(capacity)]
        curr_lines = transmission_lines[
            (transmission_lines["voltage"] > v_min)
            & (transmission_lines["voltage"] <= v_max)
        ]
        if len(curr_lines) == 0:
            continue
        out = _rasterize_transmission_layer(curr_lines, cost_shape,
                                            cost_transform)
        transmission_lines_dict[int(capacity)] = out
        v_min = v_max

    curr_lines = transmission_lines[transmission_lines["voltage"] > v_min]
    if len(curr_lines) == 0:
        return transmission_lines_dict

    out = _rasterize_transmission_layer(curr_lines, cost_shape, cost_transform)
    transmission_lines_dict[int(capacities[-1])] = out
    return transmission_lines_dict


def _rasterize_transmission_layer(transmission_lines, cost_shape,
                                  cost_transform):
    """Rasterize a single transmission layer. """
    shapes = [(geom, 1) for geom in transmission_lines["geometry"]
              if geom is not None]
    out = np.zeros(cost_shape, dtype='uint8')
    rasterio.features.rasterize(shapes=shapes, out=out, out_shape=out.shape,
                                fill=0, transform=cost_transform)

    return out


def min_reinforcement_costs(table, group_col="poi_gid"):
    """Filter table down to cheapest reinforcement per substation.

    Parameters
    ----------
    table : pd.DataFrame | gpd.GeoDataFrame
        Table containing costs for reinforced transmission. Must contain
        a `group_col` column identifying each substation with its own
        unique ID and a `reinforcement_cost_per_mw` column with the
        reinforcement costs to minimize.
    group_col : str, default="poi_gid"
        Name of column containing unique identifier for each substation.
        This column will be used to group costs and select the minimum
        one.

    Returns
    -------
    pd.DataFrame | gpd.GeoDataFrame
        Table with a single entry for each `group_col` with the least
        `reinforcement_cost_per_mw`.
    """
    logger.debug("Computing min reinforcement cost grouped by %s", group_col)
    logger.debug("Reinforcement table contains %d paths", len(table))

    logger.debug("Dropping NA reinforcement costs")
    table = table[~table["reinforcement_cost_per_mw"].isna()]
    logger.debug("Reinforcement table with no NA costs contains %d paths",
                 len(table))
    grouped = table.groupby(group_col)
    logger.debug("Reinforcement table contains %d groups of %s",
                 len(grouped), group_col)
    table = table.loc[grouped["reinforcement_cost_per_mw"].idxmin()]
    return table.reset_index(drop=True)


def _collect_future_chunks(futures, least_cost_paths):
    """Collect all futures from the input dictionary. """

    num_to_collect = len(futures)
    for i, future in enumerate(as_completed(futures), start=1):
        end_features = futures.pop(future)
        lcp = future.result()
        logger.debug("Joining feautres of shape %s with results of shape %s",
                     str(end_features.shape), str(lcp.shape))
        logger.debug("Feature cols: %s", str(end_features.columns))
        logger.debug("Results cols: %s", str(lcp.columns))
        lcp = lcp.merge(end_features, on=["row", "col"])
        least_cost_paths.append(lcp)
        logger.debug('Collected %d of %d futures!', i, num_to_collect)
        log_mem(logger)

    return least_cost_paths
